import asyncio
from typing import List, Optional

from leaf_playground.core.workers import Logger
from leaf_playground.core.scene import Scene
from leaf_playground.core.scene_definition import SceneConfig
from leaf_playground.data.log_body import ActionLogBody
from leaf_playground.data.media import Text
from pydantic import Field

from .agents.examiner import Examiner
from .agents.base_examinee import AIBaseExaminee
from .dataset_util import *
from .scene_definition import *


class MmluSceneLogBody(ActionLogBody):
    ground_truth: Optional[Text] = Field(default=None)


MmluSceneConfig = SceneConfig.create_config_model(
    SCENE_DEFINITION, additional_config_fields={"dataset_config": (DatasetConfig, Field(default=...))}
)


class MmluScene(Scene, scene_definition=SCENE_DEFINITION, log_body_class=MmluSceneLogBody):
    config_cls = MmluSceneConfig
    config: config_cls

    def __init__(self, config: config_cls):
        super().__init__(config=config)

        self.examiner: Examiner = self.static_agents["examiner"][0]
        self.examinees: List[AIBaseExaminee] = self.agents["examinee"]

    async def _run(self):
        async def examinee_answer(examinee: AIBaseExaminee, s: ExaminerSample) -> None:
            try:
                answer: ExamineeAnswer = await examinee.answer(sample=s, examiner=self.examiner.profile)
            except:
                if self.config.debug_mode:
                    raise
                answer: ExamineeAnswer = ExamineeAnswer(
                    sender=examinee.profile,
                    receivers=[self.examiner.profile],
                    content=Text(text=""),
                    sample_id=s.sample_id,
                )
            self.message_pool.put_message(answer)
            ground_truth = self.examiner.get_golden_answer(s.sample_id)
            log = self.log_body_class(
                references=[s.id],
                response=answer.id,
                ground_truth=Text(text=ground_truth) if ground_truth else None,
                log_msg=f"examinee [{examinee.name}] answers to sample [{s.sample_id}]",
                action_belonged_chain=examinee.role_definition.get_action_definition("answer").belonged_chain,
            )
            self.logger.add_log(log)
            self.notify_evaluators_record(log)

        await self.examiner.prepare_samples(self.config.dataset_config)
        while not self.examiner.check_examine_finish():
            sample: ExaminerSample = await self.examiner.send_sample(
                receivers=[examinee.profile for examinee in self.examinees]
            )
            self.message_pool.put_message(sample)
            self.logger.add_log(
                self.log_body_class(
                    references=None,
                    response=sample.id,
                    ground_truth=None,
                    log_msg=f"examiner sends sample [{sample.sample_id}] to all examinees",
                    action_belonged_chain=self.examiner.role_definition.get_action_definition(
                        "send_sample"
                    ).belonged_chain,
                )
            )

            await asyncio.gather(*[examinee_answer(examinee, sample) for examinee in self.examinees])


__all__ = ["MmluSceneConfig", "MmluScene"]



from sys import _getframe
from typing import Any, List, Literal, Optional, Tuple, Type, Union

from pydantic_core import PydanticUndefined
from pydantic import create_model, BaseModel, Field, PositiveInt, PrivateAttr

from .action import ActionConfig, ActionDefinition
from ...._config import _Config
from ....data.profile import Role
from ....utils.import_util import dynamically_import_obj, DynamicObject


class RoleDefinition(BaseModel):
    name: str = Field(default=...)
    description: str = Field(default=...)
    actions: List[ActionDefinition] = Field(default=...)
    num_agents_range: Tuple[PositiveInt, Union[PositiveInt, Literal[-1]]] = Field(default=(1, 1))
    is_static: bool = Field(default=False)

    _agents_cls: List[Type] = PrivateAttr(default=[])
    _role_instance: Optional[Role] = PrivateAttr(default=None)

    @property
    def agents_cls(self) -> List[Type["leaf_playground.core.scene_agent.SceneAgent"]]:
        return self._agents_cls

    @property
    def agents_metadata(self) -> List[BaseModel]:
        if not self.agents_cls:
            raise ValueError(f"no agent class is implemented for role [{self.name}]")
        return [cls.get_metadata() for cls in self.agents_cls]

    @property
    def min_agents_num(self) -> int:
        return self.num_agents_range[0]

    @property
    def max_agents_num(self) -> int:
        return self.num_agents_range[1]

    @property
    def role_instance(self) -> Role:
        if not self._role_instance:
            self._role_instance = Role(name=self.name, description=self.description, is_static=self.is_static)
        return self._role_instance

    def model_post_init(self, __context: Any) -> None:
        # validate num_agents_range
        if not self.is_static:
            if self.num_agents_range[1] != -1 and self.num_agents_range[1] < self.num_agents_range[0]:
                raise ValueError("the second item of num_agents_range must be -1 or greater than the first item")
        elif self.num_agents_range != (1, 1):
            raise ValueError("static role can only have one agent at a time")

        if not self.actions and not self.is_static:
            raise ValueError("dynamic role's actions can't be empty")
        if len(set([a.name for a in self.actions])) != len(self.actions):
            raise ValueError(f"actions of role [{self.name}] should have unique names")
        for action in self.actions:
            if action.belonged_role:
                raise ValueError(
                    f"[{action.name}] action has already been bounded to role [{action.belonged_role.name}]"
                )
            action._belonged_role = self
        # compare metrics can only be used in static role
        for action in self.actions:
            for metric in action.metrics or []:
                if metric.is_comparison and not self.is_static:
                    raise ValueError(
                        f"[{metric.belonged_chain}] metric is a comparison metric, can only be used for static role."
                    )

    def get_action_definition(self, action_name: str) -> ActionDefinition:
        for action in self.actions:
            if action.name == action_name:
                return action
        raise ValueError(f"action [{action_name}] not found")


class RoleActionsConfig(_Config):
    @classmethod
    def create_config_model(cls, role_definition: RoleDefinition) -> Type["RoleActionsConfig"]:
        model_name = "".join([each.capitalize() for each in role_definition.name.split("_")]) + cls.__name__
        module = _getframe(1).f_globals["__name__"]
        fields = {}
        for action in role_definition.actions:
            fields[action.name] = (ActionConfig.create_config_model(action), Field(default=...))
        return create_model(__model_name=model_name, __module__=module, __base__=cls, **fields)

    def get_action_config(self, action_name: str) -> ActionConfig:
        return getattr(self, action_name)


class RoleAgentConfig(_Config):
    config_data: dict = Field(default=...)
    obj_for_import: DynamicObject = Field(default=...)

    _agent_config = PrivateAttr(default=None)
    _agent_cls = PrivateAttr(default=None)

    @property
    def agent_config(self) -> "leaf_playground.core.scene_agent.SceneAgentConfig":
        if self._agent_config is None:
            self._agent_config = dynamically_import_obj(self.obj_for_import).config_cls(**self.config_data)
        return self._agent_config

    @property
    def agent_cls(self) -> Type["leaf_playground.core.scene_agent.SceneAgent"]:
        if self._agent_cls is None:
            self._agent_cls = dynamically_import_obj(self.obj_for_import)
        return self._agent_cls

    def initiate_agent(self) -> "leaf_playground.core.scene_agent.SceneAgent":
        return self.agent_cls(config=self.agent_config)


class RoleConfig(_Config):
    actions_config: RoleActionsConfig = Field(default=...)
    agents_config: List[RoleAgentConfig] = Field(default=...)
    is_static: bool = Field(default=...)

    _role_definition: RoleDefinition = PrivateAttr(default=None)

    def __init_subclass__(cls, _role_definition: RoleDefinition, **kwargs):
        super().__init_subclass__(**kwargs)
        cls._role_definition = _role_definition

    @property
    def role_definition(self) -> RoleDefinition:
        return self._role_definition

    def model_post_init(self, __context: Any) -> None:
        # validate agents_config
        if not self.agents_config:
            raise ValueError("agents_config can't empty")
        num_agents = len(self.agents_config)
        # make sure all agents' name is unique
        if num_agents != len(set([conf.agent_config.profile.name for conf in self.agents_config])):
            raise AttributeError(f"each agents' name should be unique")
        # make sure all agent class can be recognized by role definition
        for agent_config in self.agents_config:
            if agent_config.agent_cls.role_definition != self.role_definition:
                raise TypeError(
                    f"expected agents class for role [{self.role_definition.name}] are: "
                    f"{self.role_definition.agents_cls}, got {agent_config.agent_cls}"
                )

    @classmethod
    def create_config_model(cls, role_definition: RoleDefinition) -> "RoleConfig":
        model_name = "".join([each.capitalize() for each in role_definition.name.split("_")]) + cls.__name__
        module = _getframe(1).f_globals["__name__"]
        cls_kwargs = {"_role_definition": role_definition}
        fields = {
            "actions_config": (RoleActionsConfig.create_config_model(role_definition), Field(default=...)),
            "agents_config": (
                List[RoleAgentConfig],
                Field(
                    default=(
                        ...
                        if not role_definition.is_static
                        else [
                            RoleAgentConfig(
                                config_data={}, obj_for_import=role_definition.agents_cls[0].obj_for_import
                            )
                        ]
                    ),
                    min_items=role_definition.min_agents_num,
                    max_items=(
                        role_definition.max_agents_num if role_definition.max_agents_num != -1 else PydanticUndefined
                    ),
                ),
            ),
            "is_static": (Literal[role_definition.is_static], Field(default=role_definition.is_static)),
        }

        return create_model(
            __model_name=model_name, __module__=module, __base__=cls, __cls_kwargs__=cls_kwargs, **fields
        )

    def get_action_config(self, action_name: str) -> ActionConfig:
        return self.actions_config.get_action_config(action_name)

    def initiate_agents(self) -> List["leaf_playground.core.scene_agent.SceneAgent"]:
        return [conf.initiate_agent() for conf in self.agents_config]


__all__ = ["RoleDefinition", "RoleActionsConfig", "RoleAgentConfig", "RoleConfig"]



import inspect
import json
from typing import Literal, Type

from pydantic import create_model, BaseModel


class _Config(BaseModel):
    def save(self, file_path: str):
        with open(file_path, "w", encoding="utf-8") as f:
            json.dump(self.model_dump(mode="json", by_alias=True), f, indent=2)

    @classmethod
    def load(cls, file_path: str):
        kwargs = json.load(open(file_path, "r", encoding="utf-8"))
        return cls(**kwargs)

    @classmethod
    def get_json_schema(cls, by_alias: bool = False, mode: Literal["validation", "serialization"] = "validation"):
        def _create_new_annotation(annotation: Type):
            if not inspect.isclass(annotation) or not issubclass(annotation, BaseModel):
                return annotation
            anno_fields = {f_name: f_info for f_name, f_info in annotation.model_fields.items() if not f_info.exclude}
            for f_name, f_info in anno_fields.items():
                f_info.annotation = _create_new_annotation(f_info.annotation)
                anno_fields[f_name] = (f_info.annotation, f_info)
            return create_model(
                __model_name=f"{annotation.__name__}", __module__="pydantic_temp", __base__=BaseModel, **anno_fields
            )

        return _create_new_annotation(cls).model_json_schema(by_alias=by_alias, mode=mode)


class _Configurable:
    config_cls = _Config

    def __init__(self, config: config_cls):
        if not isinstance(config, self.config_cls):
            raise TypeError(
                f"required '{self.config_cls.__name__}' type config, but get '{config.__class__.__name__}'"
            )
        self.config = config

    def __init_subclass__(cls, **kwargs):
        if cls.config_cls == _Config or not issubclass(cls.config_cls, _Config):
            raise ValueError(
                "'_config_type' must be _Config's sub-class and can't be _Config itself, "
                "which means, when design a _Configurable class, you must also design a "
                "_Config class and override '_config_type' of the new _Configurable class "
                "with the new _Config class."
            )

    def to_dict(self, **kwargs) -> dict:
        return self.config.model_dump(**kwargs)

    def to_json(self, **kwargs) -> str:
        return self.config.model_dump_json(**kwargs)

    def save_config(self, file_path: str):
        self.config.save(file_path)

    @classmethod
    def from_config(cls, config: _Config) -> "_Configurable":
        raise NotImplementedError(f"{cls.__name__} not implements from_config method")

    @classmethod
    def from_config_file(cls, file_path: str) -> "_Configurable":
        config = cls.config_cls.load(file_path)
        return cls.from_config(config)



from abc import abstractmethod, ABC, ABCMeta
from sys import _getframe
from typing import List, Optional

from pydantic import BaseModel, Field

from ..._type import Immutable
from ..scene_definition import CombinedMetricsData, SceneConfig
from ...data.log_body import LogBody
from ...utils.import_util import DynamicObject
from ...utils.type_util import validate_type


_chart_names = set()


class ChartMetaClass(ABCMeta):
    def __new__(cls, name, bases, attrs, *, chart_name: str = None, supported_metric_names: List[str] = None):
        if chart_name in _chart_names:
            raise ValueError(f"Chart name {chart_name} has been used!")
        _chart_names.add(chart_name)
        attrs["chart_name"] = Immutable(chart_name)
        attrs["supported_metric_names"] = Immutable(supported_metric_names)
        attrs["obj_for_import"] = Immutable(DynamicObject(obj=name, module=_getframe(1).f_globals["__name__"]))

        new_cls = super().__new__(cls, name, bases, attrs)

        DynamicObject.bind_dynamic_obj(attrs["obj_for_import"], new_cls)

        if not validate_type(attrs["chart_name"], expect_type=Immutable[Optional[str]]):
            raise TypeError(f"class [{name}]'s class attribute [chart_name] should be a str")
        if not validate_type(attrs["supported_metric_names"], expect_type=Immutable[Optional[List[str]]]):
            raise TypeError(f"class [{name}]'s class attribute [chart_name] should be a List[str]")

        if ABC not in bases:
            # check if those class attrs are empty when the class is not abstract
            if not new_cls.chart_name:
                raise AttributeError(
                    f"class [{name}] missing class attribute [chart_name], please specify it by "
                    f"doing like: `class {name}(chart_name=your_chart_name)`, where 'your_chart_name' "
                    "is a string that introduces your chart class"
                )
            if not new_cls.supported_metric_names:
                raise AttributeError(
                    f"class [{name}] missing class attribute [supported_metric_names], please specify it by "
                    f"doing like: `class {name}(supported_metric_names=your_supported_metric_names)`, where "
                    "'your_supported_metric_names' is a list of string that introduces the metric names "
                    "supported by your chart class"
                )
        return new_cls

    def __init__(cls, name, bases, attrs, *, chart_name: str = None, supported_metric_names: List[str] = None):
        super().__init__(name, bases, attrs)


class ChartMetadata(BaseModel):
    cls_name: str = Field(default=...)
    obj_for_import: DynamicObject = Field(default=...)
    chart_name: str = Field(default=...)
    supported_metric_names: List[str] = Field(default=...)


class Chart(ABC, metaclass=ChartMetaClass):
    obj_for_import: DynamicObject
    supported_metric_names: List[str]
    chart_name: str

    @classmethod
    def get_metadata(cls) -> ChartMetadata:
        return ChartMetadata(
            cls_name=cls.__name__,
            obj_for_import=cls.obj_for_import,
            chart_name=cls.chart_name,
            supported_metric_names=cls.supported_metric_names,
        )

    def generate(
        self,
        metrics: CombinedMetricsData,
        scene_config: SceneConfig,
        evaluator_configs: List["leaf_playground.core.workers.MetricEvaluatorConfig"],
        logs: List[LogBody],
    ) -> Optional[dict]:
        if not metrics:
            return None

        filtered_metrics: CombinedMetricsData = {
            "metrics": {k: v for k, v in metrics["metrics"].items() if k in self.supported_metric_names},
            "human_metrics": {k: v for k, v in metrics["human_metrics"].items() if k in self.supported_metric_names},
            "merged_metrics": {k: v for k, v in metrics["merged_metrics"].items() if k in self.supported_metric_names},
        }
        try:
            return self._generate(filtered_metrics, scene_config, evaluator_configs, logs)
        except Exception as e:
            print(f"Error occurred when generating chart {self.__class__.__name__}: {e}")
            return None

    @abstractmethod
    def _generate(
        self,
        metrics: CombinedMetricsData,
        scene_config: SceneConfig,
        evaluator_configs: List["leaf_playground.core.workers.MetricEvaluatorConfig"],
        logs: List[LogBody],
    ) -> dict:
        pass


__all__ = ["Chart", "ChartMetadata"]



from leaf_playground.core.scene_definition import *


SCENE_DEFINITION = SceneDefinition(
    name="{{ cookiecutter.project_name_camel_case }}",
    description="",
    roles=[],
    env_vars=[]
)


__all__ = [
    "SCENE_DEFINITION"
]



from enum import Enum
from typing import Literal

from pydantic import Field

from .base import Data


class SocketOperation(Enum):
    CREATE = "create"
    UPDATE = "update"


class SocketData(Data):
    data: dict = Field(default=...)
    operation: SocketOperation = Field(default=SocketOperation.CREATE)
    type: Literal["socket_data"] = Field(default="data")


class SocketEvent(Data):
    event: str = Field(default=...)
    type: Literal["socket_event"] = Field(default="event")


__all__ = ["SocketData", "SocketEvent", "SocketOperation"]



from abc import abstractmethod, ABC

from leaf_playground.core.scene_agent import SceneAIAgentConfig, SceneAIAgent
from leaf_playground.data.profile import Profile

from ..scene_definition import ExamineeAnswer, ExaminerSample, SCENE_DEFINITION


ROLE_DEFINITION = SCENE_DEFINITION.get_role_definition("examinee")


class ExamineeConfig(SceneAIAgentConfig):
    pass


class Examinee(SceneAIAgent, ABC, role_definition=ROLE_DEFINITION):
    config_cls = ExamineeConfig
    config: config_cls

    @abstractmethod
    async def answer(self, sample: ExaminerSample, examiner: Profile) -> ExamineeAnswer:
        pass


__all__ = ["ExamineeConfig", "Examinee"]



import base64
import json
import os
import traceback
import warnings
from typing import Dict, List, Optional, Any

import aiohttp
from fastapi import status, APIRouter, Depends, HTTPException
from fastapi.responses import JSONResponse, PlainTextResponse
from pydantic import BaseModel, Field, PrivateAttr

from leaf_playground._type import Singleton


hub_router = APIRouter(prefix="/hub")


class ProjectMetadata(BaseModel):
    scene_metadata: dict = Field(default=dict())
    agents_metadata: Dict[str, List[dict]] = Field(default=dict())
    evaluators_metadata: Optional[List[dict]] = Field(default=None)
    charts_metadata: Optional[List[dict]] = Field(default=None)


class Project(BaseModel):
    # fields load from project config
    name: str = Field(default=...)
    id: str = Field(default=...)
    version: str = Field(default=...)
    leaf_version: str = Field(default=...)
    created_at: str = Field(default=...)
    last_update: str = Field(default=...)
    metadata: ProjectMetadata = Field(default=...)

    # fields set in run time
    work_dir: str = Field(default=...)
    readme: str = Field(default="")
    requirements: str = Field(default="")
    dockerfile: str = Field(default="")

    def model_post_init(self, __context: Any) -> None:
        self.readme = self._get_readme()
        self.requirements = self._get_requirements()
        self.dockerfile = self._get_dockerfile()

    @property
    def has_readme(self) -> bool:
        return bool(self.readme)

    @property
    def has_requirements(self) -> bool:
        return bool(self.requirements)

    @property
    def has_dockerfile(self) -> bool:
        return bool(self.dockerfile)

    @staticmethod
    def _read_file(fpath) -> str:
        if not os.path.exists(fpath):
            return ""
        with open(fpath, "r", encoding="utf-8") as f:
            content = f.read()
        return content

    def _get_readme(self) -> str:
        return self._read_file(os.path.join(self.work_dir, "README.md"))

    def _get_requirements(self) -> str:
        return self._read_file(os.path.join(self.work_dir, "requirements.txt"))

    def _get_dockerfile(self) -> str:
        return self._read_file(os.path.join(self.work_dir, "Dockerfile"))

    def update_readme(self):
        self.readme = self._get_readme()

    def update_requirements(self):
        self.requirements = self._get_requirements()

    def update_dockerfile(self):
        self.dockerfile = self._get_dockerfile()

    def update_config(self):
        with open(os.path.join(self.work_dir, ".leaf", "project_config.json"), "r", encoding="utf-8") as f:
            proj_config = json.load(f)
        for k, v in proj_config.items():
            if k in self.model_fields_set:
                if k != "metadata":
                    setattr(self, k, v)
                else:
                    setattr(self, k, ProjectMetadata(**v))

    def update(self):
        self.update_config()
        self.update_readme()
        self.update_requirements()
        self.update_dockerfile()


class Hub(Singleton):
    def __init__(self, hub_dir: str):
        self._hub_dir = hub_dir

        self._projects: Dict[str, Project] = {}
        self.scan_projects()

    @property
    def hub_dir(self) -> str:
        return self._hub_dir

    @property
    def projects(self) -> Dict[str, Project]:
        return self._projects

    def scan_projects(self) -> None:
        def _scan_projects(hub_dir: str) -> List[Project]:
            projects_list = []

            if not os.path.isdir(hub_dir) or not os.path.exists(hub_dir):
                return projects_list

            if os.path.exists(os.path.join(hub_dir, ".leaf")):
                work_dir = hub_dir
                with open(os.path.join(work_dir, ".leaf", "project_config.json"), "r", encoding="utf-8") as f:
                    try:
                        project = Project(**json.load(f), work_dir=work_dir)
                    except:
                        traceback.print_exc()
                        warnings.warn(
                            f"load project config from {os.path.join(work_dir, '.leaf')} failed, "
                            "will ignore this project."
                        )
                    else:
                        project.work_dir = work_dir
                        projects_list.append(project)
            else:
                for root, dirs, _ in os.walk(hub_dir):
                    for work_dir in dirs:
                        projects_list += _scan_projects(str(os.path.join(root, work_dir)))

            return projects_list

        self._projects = {proj.id: proj for proj in _scan_projects(self._hub_dir)}

    def get_project(self, project_id: str) -> Project:
        try:
            project = self._projects[project_id]
        except KeyError:
            raise KeyError(f"project [{project_id}] not found, make sure you have this project under {self._hub_dir}")
        return project

    @classmethod
    def http_get_project_by_id(cls, project_id: str) -> Project:
        try:
            hub = cls.get_instance()
        except KeyError as e:
            raise HTTPException(status_code=status.HTTP_503_SERVICE_UNAVAILABLE, detail="service not ready.")

        try:
            project = hub.get_project(project_id)
        except KeyError as e:
            raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail=str(e))

        return project


@hub_router.get("/refresh", response_class=JSONResponse)
async def refresh_projects(hub: Hub = Depends(Hub.get_instance)) -> JSONResponse:
    """refresh projects, this will re-scan projects under the hub dir, respond a list of updated project id"""

    hub.scan_projects()
    return JSONResponse(content=list(hub.projects.keys()))


@hub_router.get("/{project_id}/info", response_model=Project)
async def get_project_info(project: Project = Depends(Hub.http_get_project_by_id)) -> Project:
    """retrieve a project's information by giving its id, this will reload local files to update first."""
    project.update()
    return project


@hub_router.get("/{project_id}/readme.md", response_class=PlainTextResponse)
async def get_project_readme(project: Project = Depends(Hub.http_get_project_by_id)) -> PlainTextResponse:
    """retrieve a project's README.md by giving its id, this will reload local file to update first"""
    project.update_readme()
    return PlainTextResponse(content=project.readme)


@hub_router.get("/{project_id}/requirements.txt", response_class=PlainTextResponse)
async def get_project_requirements(project: Project = Depends(Hub.http_get_project_by_id)) -> PlainTextResponse:
    """retrieve a project's requirements.txt by giving its id, this will reload local file to update first"""
    project.update_requirements()
    return PlainTextResponse(content=project.requirements)


@hub_router.get("/{project_id}/dockerfile", response_class=PlainTextResponse)
async def get_project_dockerfile(project: Project = Depends(Hub.http_get_project_by_id)) -> PlainTextResponse:
    """retrieve a project's Dockerfile by giving its id, this will reload local file to update first"""
    project.update_dockerfile()
    return PlainTextResponse(content=project.dockerfile)


def validate_file_path(file_path: str):
    if file_path.startswith(".") and not file_path.startswith("./"):
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY, detail="relevant path must starts with './'"
        )
    return file_path


@hub_router.get("/{project_id}/assets/download_image", response_class=JSONResponse)
async def get_project_image_asset(
    project: Project = Depends(Hub.http_get_project_by_id), file_path: str = Depends(validate_file_path)
) -> JSONResponse:
    """
    retrieve a project's image asset by giving project_id and file_path, support both local path and http url,
    if given local path is a relevant path, it must start with './', otherwise responds with 422 status code
    """
    if file_path.startswith(("http://", "https://")):
        async with aiohttp.ClientSession() as session:
            async with session.get(file_path) as response:
                if response.status == 200:
                    image_data = await response.read()
                else:
                    raise HTTPException(
                        status_code=status.HTTP_404_NOT_FOUND, detail=f"can't connect to URL [{file_path}]."
                    )
    else:
        if file_path.startswith("."):
            file_path = os.path.join(project.work_dir, file_path[2:])
        if not os.path.isfile(file_path):
            raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail=f"local file [{file_path}] not found.")
        with open(file_path, "rb") as f:
            image_data = f.read()

    return JSONResponse(content={"image": base64.b64encode(image_data).decode()})


# TODO: support download audio and video


__all__ = ["hub_router", "ProjectMetadata", "Project", "Hub"]



import glob
import importlib
import importlib.machinery
import inspect
import os
import pkgutil
from functools import partial
from hashlib import md5
from pathlib import Path
from typing import Any, List, Optional, Type

from pydantic import BaseModel, Field, FilePath

_IMPORTED_OBJECTS = {}
_IMPORTED_FUNCTIONS = {}


class DynamicObject(BaseModel):
    obj: str = Field(default=...)
    module: Optional[str] = Field(default=None)
    source_file: Optional[FilePath] = Field(default=None)

    def model_post_init(self, __context: Any) -> None:
        if self.module is None and self.source_file is None:
            raise ValueError("at least one of 'module' or 'source_file' should not None.")
        if self.module and self.source_file:
            raise ValueError("can't specify 'module' and 'source_file' at the same time.")

    @property
    def hash(self) -> str:
        obj = self.obj
        module = "" if not self.module else self.module
        source_file = "" if not self.source_file else self.source_file.as_posix()
        return md5((obj + module + source_file).encode(encoding="utf-8")).hexdigest()

    @classmethod
    def create_dynamic_obj(cls, obj: Type) -> "DynamicObject":
        dynamic_obj = cls(obj=obj.__name__, module=inspect.getmodule(obj).__name__)
        _IMPORTED_OBJECTS[dynamic_obj.hash] = obj
        return dynamic_obj

    @classmethod
    def bind_dynamic_obj(cls, obj: "DynamicObject", target: Type) -> None:
        _IMPORTED_OBJECTS[obj.hash] = target


def dynamically_import_obj(o: DynamicObject):
    if o.hash in _IMPORTED_OBJECTS:
        return _IMPORTED_OBJECTS[o.hash]
    if o.module is not None:
        module = importlib.import_module(o.module)
    else:
        module = importlib.machinery.SourceFileLoader(o.source_file.name, o.source_file.as_posix()).load_module()
    obj = module.__dict__[o.obj]
    _IMPORTED_OBJECTS[o.hash] = obj
    return obj


class DynamicFn(BaseModel):
    fn: DynamicObject = Field(default=...)
    default_kwargs: Optional[dict] = Field(default=None)

    @property
    def hash(self) -> str:
        return self.fn.hash + md5(str(self.default_kwargs).encode(encoding="utf-8")).hexdigest()

    @classmethod
    def create_dynamic_fn(cls, fn: Type, default_kwargs: Optional[dict] = None) -> "DynamicFn":
        dynamic_obj = DynamicObject.create_dynamic_obj(fn)
        dynamic_fn = cls(fn=dynamic_obj, default_kwargs=default_kwargs)
        _IMPORTED_FUNCTIONS[dynamic_fn.hash] = fn
        return dynamic_fn


def dynamically_import_fn(f: DynamicFn):
    if f.hash in _IMPORTED_FUNCTIONS:
        return _IMPORTED_FUNCTIONS[f.hash]
    fn = dynamically_import_obj(f.fn)
    if f.default_kwargs is not None:
        fn = partial(fn, **f.default_kwargs)
    _IMPORTED_FUNCTIONS[f.hash] = fn
    return fn


def find_subclasses(package_path: str, base_class: Type) -> List[DynamicObject]:
    classes = []

    for file in glob.glob(os.path.join(package_path, f"**/*.py"), recursive=True):
        module_path = Path(file)
        module = importlib.machinery.SourceFileLoader(module_path.name, module_path.as_posix()).load_module()

        if not hasattr(module, "__all__"):
            continue
        # TODO: figure out why using module.__dict__ directly will get duplicate objects
        for name in module.__all__:
            obj = module.__dict__[name]
            if not inspect.isclass(obj):
                continue
            if issubclass(obj, base_class) and obj != base_class and obj.__module__ == module.__name__:
                classes.append(DynamicObject(obj=obj.__name__, source_file=module_path))

    return classes


def relevantly_find_subclasses(root_path: str, prefix: str, base_class: type) -> List[Type]:
    classes = []
    for loader, module_name, is_pkg in pkgutil.walk_packages(path=[root_path]):
        if is_pkg:
            classes += relevantly_find_subclasses(
                os.path.join(root_path, module_name), prefix=prefix + "." + module_name, base_class=base_class
            )
        else:
            module = importlib.import_module(prefix + "." + module_name)
            for name in module.__all__:
                obj = module.__dict__[name]
                if not inspect.isclass(obj):
                    continue
                if issubclass(obj, base_class) and obj != base_class and obj.__module__ == module.__name__:
                    classes.append(obj)
    return classes


__all__ = [
    "DynamicObject",
    "DynamicFn",
    "dynamically_import_obj",
    "dynamically_import_fn",
    "find_subclasses",
    "relevantly_find_subclasses",
]



from datetime import datetime
from typing import Dict, List, Literal, Union
from uuid import uuid4, UUID

from pydantic import Field
from typing_extensions import Annotated

from .base import Data
from .media import Media, Text, Json, Image, Audio, Video
from .profile import Profile
from .._type import Singleton


class Message(Data):
    id_: UUID = Field(default_factory=uuid4)
    sender: Profile = Field(default=...)
    content: Media = Field(default=...)
    receivers: List[Profile] = Field(default=...)
    msg_type: Literal["basic"] = Field(default="basic")
    created_at: datetime = Field(default_factory=lambda: datetime.utcnow())

    @property
    def id(self):
        return self.sender_id + "_" + self.id_.hex

    @property
    def sender_name(self):
        return self.sender.name

    @property
    def sender_id(self):
        return self.sender.id

    @property
    def sender_role(self):
        return self.sender.role.name

    @property
    def receiver_names(self):
        return [receiver.name for receiver in self.receivers]

    @property
    def receiver_ids(self):
        return [receiver.id for receiver in self.receivers]

    @property
    def receiver_roles(self):
        return list(set([receiver.role.name for receiver in self.receivers]))


class TextMessage(Message):
    content: Text = Field(default=...)
    msg_type: Literal["text"] = Field(default="text")


class JsonMessage(Message):
    content: Json = Field(default=...)
    msg_type: Literal["json"] = Field(default="json")


class ImageMessage(Message):
    content: Image = Field(default=...)
    msg_type: Literal["image"] = Field(default="image")


class AudioMessage(Message):
    content: Audio = Field(default=...)
    msg_type: Literal["audio"] = Field(default="audio")


class VideoMessage(Message):
    content: Video = Field(default=...)
    msg_type: Literal["video"] = Field(default="video")


class MessagePool(Singleton):
    """
    A global message pool to cache all agents' messages in one engine run

    :param messages: messages that all agents send in one engine run
    :type messages: Dict[str, Message]
    """

    # TODO: fix: design message pool as a singleton and allow get_instance globally may let agents to cheat
    def __init__(self):
        self.messages: Dict[str, Message] = {}

    def clear(self):
        """Clear cached messages, at the start of each engine run, this method should be called"""
        self.messages = {}

    def put_message(self, message: Message):
        """Put one message into the cache"""
        self.messages[message.id] = message

    def get_messages(self, agent: Profile) -> List[Message]:
        """
        Get messages that sent by the agent or is visible by the agent

        :param agent: agent who want to retrieve messages
        :type agent: Profile
        :return: a list of messages that are sent by or visible by the agent
        :rtype: List[Message]
        """
        messages = []
        for message in self.messages.values():
            receivers_ids = [receiver.id for receiver in message.receivers]
            if agent.id in receivers_ids:
                messages.append(message)
        return messages

    def get_message_by_id(self, msg_id: str) -> Message:
        return self.messages[msg_id]


BasicMessageType = Annotated[
    Union[TextMessage, JsonMessage, ImageMessage, AudioMessage, VideoMessage], Field(discriminator="msg_type")
]

__all__ = [
    "Message",
    "TextMessage",
    "JsonMessage",
    "ImageMessage",
    "AudioMessage",
    "VideoMessage",
    "MessagePool",
    "BasicMessageType",
]



import asyncio
import json
import warnings
import os
from abc import abstractmethod
from typing import Dict, List, Literal, Any

import pandas as pd
from pydantic import BaseModel, Field

from ...data.log_body import LogBody, ActionLogBody
from ...data.message import MessagePool
from ...utils.thread_util import run_asynchronously
from ..._type import Singleton


class Logger(Singleton):
    def __init__(self):
        self.message_pool: MessagePool = MessagePool.get_instance()

        self._logs: Dict[str, LogBody] = {}
        self._stopped: bool = False
        self._handlers: List["LogHandler"] = []

    def registry_handler(self, handler: "LogHandler"):
        self._handlers.append(handler)

    def add_log(self, log_body: LogBody):
        self._logs[log_body.id] = log_body

        for handler in self._handlers:
            asyncio.ensure_future(run_asynchronously(handler.notify_create, log_body))

    def add_action_log_record(
        self,
        log_id: str,
        records: Dict[str, dict],
        field_name: Literal["eval_records", "compare_records", "human_eval_records", "human_compare_records"],
    ):
        log_body = self._logs[log_id]
        for name, record in records.items():
            if "human" not in field_name:
                getattr(log_body, field_name)[name].append(record)
            else:
                getattr(log_body, field_name)[name] = record

        for handler in self._handlers:
            asyncio.ensure_future(run_asynchronously(handler.notify_update, log_body))

    @property
    def logs(self) -> List[LogBody]:
        return list(self._logs.values())

    def is_log_exists(self, log_id: str) -> bool:
        return log_id in self._logs

    def get_log_by_id(self, log_id: str) -> LogBody:
        return self._logs[log_id]


class LogHandler:
    @abstractmethod
    async def notify_create(self, log: LogBody):
        pass

    @abstractmethod
    async def notify_update(self, log: LogBody):
        pass


_KEPT_LOG_FILE_NAME = ".log"


class LogExporter(BaseModel):
    file_name: str = Field(default="log")
    extension: Literal["json", "jsonl", "csv"] = Field(default="jsonl")

    def model_post_init(self, __context: Any) -> None:
        if self.file_name in _KEPT_LOG_FILE_NAME:
            raise ValueError(f"can't set file_name to {_KEPT_LOG_FILE_NAME}")
        if self.extension not in ["json", "jsonl", "csv"]:
            raise NotImplementedError(f"file extension {self.extension} isn't support yet.")

    def _export_to_json(self, logger: Logger, save_path: str):
        logs = []
        for log in logger.logs:
            log_dict = log.model_dump(mode="json")
            if isinstance(log, ActionLogBody):
                if log.references is not None:
                    log_dict["references"] = (
                        [logger.message_pool.get_message_by_id(ref).model_dump(mode="json") for ref in log.references]
                        if log.references
                        else None
                    )
                log_dict["response"] = logger.message_pool.get_message_by_id(log.response).model_dump(mode="json")
            logs.append(log_dict)

        with open(save_path, "w", encoding="utf-8") as f:
            json.dump(logs, f, indent=4, ensure_ascii=False)

    def _export_to_jsonl(self, logger: Logger, save_path: str):
        with open(save_path, "w", encoding="utf-8") as f:
            for log in logger.logs:
                log_dict = log.model_dump(mode="json")
                if isinstance(log, ActionLogBody):
                    if log.references is not None:
                        log_dict["references"] = (
                            [
                                logger.message_pool.get_message_by_id(ref).model_dump(mode="json")
                                for ref in log.references
                            ]
                            if log.references
                            else None
                        )
                    log_dict["response"] = logger.message_pool.get_message_by_id(log.response).model_dump(mode="json")
                f.write(json.dumps(log_dict) + "\n")

    def _export_to_csv(self, logger: Logger, save_path: str):
        data = {"log_id": [], "sender": [], "receivers": [], "message": []}
        for log in [log for log in logger.logs if isinstance(log, ActionLogBody)]:
            response = logger.message_pool.get_message_by_id(log.response)
            data["log_id"].append(log.id)
            data["sender"].append(response.sender_name)
            data["receivers"].append([receiver.name for receiver in response.receivers])
            data["message"].append(response.content.display_text)
        pd.DataFrame(data).to_csv(save_path, encoding="utf-8", index=False)

    def export(self, logger: Logger, save_dir: str):
        save_path = os.path.join(save_dir, f"{self.file_name}.{self.extension}")
        func_name = f"_export_to_{self.extension}"
        try:
            getattr(self, func_name)(logger, save_path)
        except:
            if os.path.exists(save_path):
                os.remove(save_path)
            warnings.warn(f"{self.__class__.__name__}.{func_name} export log failed.")


class _KeptLogExporter(LogExporter):
    file_name: str = Field(default=_KEPT_LOG_FILE_NAME)

    def model_post_init(self, __context: Any) -> None:
        pass


__all__ = ["Logger", "LogHandler", "LogExporter"]



import asyncio
from abc import abstractmethod, ABC, ABCMeta
from sys import _getframe
from typing import Dict, List, Optional, Type

from pydantic import BaseModel, Field

from .scene_agent import SceneAgent, SceneDynamicAgent, SceneHumanAgent
from .scene_definition import SceneDefinition, SceneConfig
from .workers import MetricEvaluator, Logger
from .._config import _Configurable
from .._type import Immutable
from ..data.environment import EnvironmentVariable
from ..data.log_body import ActionLogBody
from ..data.message import MessagePool
from ..utils.import_util import DynamicObject
from ..utils.type_util import validate_type


class SceneMetaClass(ABCMeta):
    def __new__(
        cls,
        name,
        bases,
        attrs,
        *,
        scene_definition: SceneDefinition = None,
        log_body_class: Type[ActionLogBody] = ActionLogBody,
    ):
        attrs["scene_definition"] = Immutable(scene_definition or getattr(bases[0], "scene_definition", None))
        attrs["log_body_class"] = Immutable(log_body_class)
        attrs["obj_for_import"] = Immutable(DynamicObject(obj=name, module=_getframe(1).f_globals["__name__"]))

        new_cls = super().__new__(cls, name, bases, attrs)

        DynamicObject.bind_dynamic_obj(attrs["obj_for_import"], new_cls)

        if not validate_type(attrs["scene_definition"], Immutable[Optional[SceneDefinition]]):
            raise TypeError(
                f"class [{name}]'s class attribute [scene_definition] should be a [SceneDefinition] instance, "
                f"got [{type(attrs['scene_definition']).__name__}] type"
            )
        if not validate_type(attrs["log_body_class"], Immutable[Type[ActionLogBody]]):
            raise TypeError(f"class [{name}]'s class attribute [log_body_class] should be subclass of [ActionLogBody]")

        if ABC not in bases:
            # check if those class attrs are empty when the class is not abstract
            if not new_cls.scene_definition:
                raise AttributeError(
                    f"class [{name}] missing class attribute [scene_definition], please specify it by "
                    f"doing like: `class {name}(scene_definition=your_scene_def)`, or you can also "
                    f"specify in the super class [{bases[0].__name__}]"
                )

        return new_cls

    def __init__(
        cls,
        name,
        bases,
        attrs,
        *,
        scene_definition: SceneDefinition = None,
        log_body_class: Type[ActionLogBody] = ActionLogBody,
    ):
        super().__init__(name, bases, attrs)

    def __setattr__(self, key, value):
        # make sure those class attributes immutable in class-wise
        if key in ["scene_definition", "log_body_class", "obj_for_import"] and hasattr(self, key):
            raise AttributeError(f"class attribute {key} is immutable")
        return super().__setattr__(key, value)

    def __delattr__(self, item):
        # make sure those class attributes can't be deleted
        if item in ["scene_definition", "log_body_class", "obj_for_import"] and hasattr(self, item):
            raise AttributeError(f"class attribute [{item}] can't be deleted")
        return super().__delattr__(item)


class SceneMetadata(BaseModel):
    scene_definition: SceneDefinition = Field(default=...)
    config_schema: dict = Field(default=...)
    obj_for_import: DynamicObject = Field(default=...)


class Scene(_Configurable, ABC, metaclass=SceneMetaClass):
    config_cls = SceneConfig
    config: SceneConfig

    # class attributes initialized in metaclass
    scene_definition: SceneDefinition
    log_body_class: Type[ActionLogBody]
    obj_for_import: DynamicObject

    def __init__(self, config: config_cls):
        super().__init__(config=config)

        static_roles = [role_def.name for role_def in self.scene_definition.roles if role_def.is_static]
        dynamic_roles = [role_def.name for role_def in self.scene_definition.roles if not role_def.is_static]
        agents: Dict[str, List[SceneAgent]] = self.config.init_agents()
        self.static_agents = {role_name: agents[role_name] for role_name in static_roles}
        self.agents = {role_name: agents[role_name] for role_name in dynamic_roles}
        self.env_vars: Dict[str, EnvironmentVariable] = self.config.init_env_vars()
        self._bind_env_vars_to_agents()
        self.evaluators: List[MetricEvaluator] = []

        self.logger = Logger.get_instance()
        self.message_pool = MessagePool.get_instance()

        self._dynamic_agents: Dict[str, SceneDynamicAgent] = {}
        self._agent_list: List[SceneAgent] = []
        self._human_agents: List[SceneHumanAgent] = []
        for agents_ in self.agents.values():
            self._human_agents += [agent for agent in agents_ if isinstance(agent, SceneHumanAgent)]
            self._agent_list += agents_
            self._dynamic_agents.update(**{agent.id: agent for agent in agents_})
        for agents_ in self.static_agents.values():
            self._agent_list += agents_

        self._run_task: Optional[asyncio.Task] = None

    @property
    def dynamic_agents(self) -> Dict[str, SceneDynamicAgent]:
        return self._dynamic_agents

    @property
    def human_agents(self) -> List[SceneHumanAgent]:
        return self._human_agents

    def get_dynamic_agent(self, agent_id: str) -> SceneDynamicAgent:
        return self._dynamic_agents[agent_id]

    def _bind_env_vars_to_agents(self):
        for agents in self.static_agents.values():
            for agent in agents:
                agent.bind_env_vars(self.env_vars)
        for agents in self.agents.values():
            for agent in agents:
                agent.bind_env_vars(self.env_vars)

    async def wait_agents_ready(self):
        while True:
            all_agents_ready = True
            for agents in self.agents.values():
                agents: List[SceneDynamicAgent]
                if not all(agent.connected for agent in agents):
                    all_agents_ready = False
            if all_agents_ready:
                break
            await asyncio.sleep(0.001)

    def registry_metric_evaluator(self, evaluator: MetricEvaluator):
        self.evaluators.append(evaluator)

    def notify_evaluators_record(self, log: ActionLogBody):
        for evaluator in self.evaluators:
            evaluator.notify_to_record(log)

    def notify_evaluators_compare(self, log: ActionLogBody):
        for evaluator in self.evaluators:
            evaluator.notify_to_compare(log)

    def notify_evaluators_can_stop(self):
        for evaluator in self.evaluators:
            evaluator.notify_can_stop()

    @abstractmethod
    async def _run(self):
        pass

    async def run(self):
        self._run_task = asyncio.ensure_future(self._run())
        try:
            await self._run_task
            self.notify_evaluators_can_stop()
        except:
            self.notify_evaluators_can_stop()
            raise

    def pause(self):
        for agent in self._agent_list:
            agent.pause()

    def resume(self):
        for agent in self._agent_list:
            agent.resume()

    def interrupt(self):
        if self._run_task and not self._run_task.done():
            self._run_task.cancel()

    @classmethod
    def get_metadata(cls) -> SceneMetadata:
        return SceneMetadata(
            scene_definition=cls.scene_definition,
            config_schema=cls.config_cls.get_json_schema(by_alias=True),
            obj_for_import=cls.obj_for_import,
        )

    @classmethod
    def from_config(cls, config: config_cls) -> "Scene":
        raise NotImplementedError()

    @classmethod
    def from_config_file(cls, file_path: str) -> "Scene":
        raise NotImplementedError()


__all__ = ["Scene", "SceneMetadata"]



from collections import defaultdict
from datetime import datetime
from enum import Enum
from typing import Dict, List, Literal, Optional, Any
from uuid import uuid4, UUID

from pydantic import model_validator, Field

from .base import Data
from .media import Media


_MetricName = str
_MetricRecord = dict
_MetricRecords = List[_MetricRecord]
_MessageID = str


class LogType(Enum):
    ACTION = "action"
    SYSTEM = "system"


class LogBody(Data):
    id: str = Field(default_factory=lambda: "log_" + uuid4().hex[:8])
    created_at: datetime = Field(default_factory=datetime.utcnow)
    last_update: datetime = Field(default=None)
    log_type: LogType = Field(default=...)
    log_msg: str = Field(default=...)

    def model_post_init(self, __context: Any) -> None:
        if self.last_update is None:
            self.last_update = self.created_at

    @model_validator(mode="before")
    def set_log_type(cls, values):
        log_type = values.get("log_type", None)
        if isinstance(log_type, str):
            log_type = log_type.lower()
            if log_type == "action":
                log_type = LogType.ACTION
            if log_type == "system":
                log_type = LogType.SYSTEM
            values["log_type"] = log_type
        return values


class ActionLogBody(LogBody):
    log_type: Literal[LogType.ACTION] = Field(default=LogType.ACTION)
    references: Optional[List[_MessageID]] = Field(default=None)
    response: _MessageID = Field(default=...)
    action_belonged_chain: Optional[str] = Field(default=...)
    ground_truth: Optional[Media] = Field(default=None)
    eval_records: Dict[_MetricName, _MetricRecords] = Field(default=defaultdict(list))
    compare_records: Dict[_MetricName, _MetricRecords] = Field(default=defaultdict(list))
    human_eval_records: Dict[_MetricName, _MetricRecord] = Field(default={})
    human_compare_records: Dict[_MetricName, _MetricRecord] = Field(default={})


class SystemEvent(Enum):
    SIMULATION_START = "simulation_start"
    SIMULATION_PAUSED = "simulation_paused"
    SIMULATION_RESUME = "simulation_resume"
    SIMULATION_FAILED = "simulation_failed"
    SIMULATION_INTERRUPTED = "simulation_interrupted"
    SIMULATION_FINISHED = "simulation_finished"
    EVALUATION_FINISHED = "evaluation_finished"
    EVERYTHING_DONE = "everything_done"


class SystemLogBody(LogBody):
    log_type: Literal[LogType.SYSTEM] = Field(default=LogType.SYSTEM)
    log_msg: Optional[str] = Field(default=None)
    system_event: SystemEvent = Field(default=...)

    def model_post_init(self, __context: Any) -> None:
        super().model_post_init(__context)
        if not self.log_msg:
            self.log_msg = self.system_event.value


__all__ = ["LogBody", "LogType", "ActionLogBody", "SystemLogBody", "SystemEvent"]



import json
import os
import re

import click
import requests
import shutil
import subprocess
import sys
import zipfile
from abc import ABC
from collections import defaultdict
from datetime import datetime
from packaging import version
from pathlib import Path
from typing import List, Optional, Type
from typing_extensions import Annotated
from urllib.request import urlopen
from uuid import uuid4

import typer
from cookiecutter.main import cookiecutter
from rich.progress import wrap_file

from leaf_playground import __version__ as leaf_version
from leaf_playground.core.scene import Scene
from leaf_playground.core.scene_agent import SceneAgent
from leaf_playground.core.workers import MetricEvaluator
from leaf_playground.core.workers.chart import Chart
from leaf_playground.utils.import_util import relevantly_find_subclasses

app = typer.Typer(name="leaf-playground-cli")
template_dir = os.path.join(os.path.dirname(__file__), "templates")


@app.command(name="new-project", help="initialize a new leaf-playground scenario simulation project from template.")
def create_new_project(name: Annotated[str, typer.Argument(metavar="project_name")]):
    project_name = name.lower().replace(" ", "_").replace("-", "_")
    project_id = project_name + "_" + uuid4().hex[:8]
    created = datetime.utcnow().strftime("%Y-%m-%d-%H-%M-%S")
    cookiecutter(
        template=template_dir,
        extra_context={
            "project_name": project_name,
            "project_id": project_id,
            "project_name_camel_case": "".join([each.capitalize() for each in project_name.split("_")]),
            "leaf_version": leaf_version,
            "created_at": created,
            "last_update": created,
        },
        no_input=True,
    )

    print(f"project [{name}] created.")
    raise typer.Exit()


@app.command(name="complete-project", help="using gpt-4 to roughly complete project code")
def complete_project(
    target: Annotated[str, typer.Option(metavar="target_dir")],
    reference: Annotated[Optional[str], typer.Option(metavar="reference_dir")] = None,
    api_key: Annotated[Optional[str], typer.Option(metavar="openai_api_key")] = None,
    disable_definition_completion: Annotated[bool, typer.Option()] = False,
    disable_agent_completion: Annotated[bool, typer.Option()] = False,
    disable_scene_completion: Annotated[bool, typer.Option()] = False,
):
    from .project_completion import Pipeline, PipelineConfig

    config = PipelineConfig(
        target_project=target,
        reference_project=reference,
        openai_api_key=api_key,
        definition_completion=not disable_definition_completion,
        agent_completion=not disable_agent_completion,
        scene_completion=not disable_scene_completion,
    )
    pipeline = Pipeline(config)
    pipeline.run()


@app.command(name="update-project-structure", help="sync project structure to the template.")
def update_project_structure(target: Annotated[str, typer.Argument(metavar="target_dir")]):
    target = os.path.abspath(target)
    dot_leaf_dir = os.path.join(target, ".leaf")
    if not os.path.exists(dot_leaf_dir):
        raise typer.BadParameter("not a leaf playground project.")
    with open(os.path.join(dot_leaf_dir, "project_config.json"), "r", encoding="utf-8") as f:
        project_config = json.load(f)
    project_name = project_config["name"]
    project_id = project_config.get("id", project_name + "_" + uuid4().hex[:8])
    created_at = project_config.get("created_at", "unknown")
    last_update = datetime.utcnow().strftime("%Y-%m-%d-%H-%M-%S")
    cookiecutter(
        output_dir=os.path.dirname(target),
        template=template_dir,
        extra_context={
            "project_name": project_name,
            "project_id": project_id,
            "project_name_camel_case": "".join([each.capitalize() for each in project_name.split("_")]),
            "leaf_version": leaf_version,
            "created_at": created_at,
            "last_update": last_update,
        },
        no_input=True,
        overwrite_if_exists=True,
        skip_if_file_exists=True,
        keep_project_on_failure=True,
    )

    # TODO: update project_config if there is update on field level

    publish_project(target, project_config["version"])

    print(f"project [{project_name}] updated.")
    raise typer.Exit()


def _replace_version_in_dockerfile(file_path, new_version):
    if not os.path.exists(file_path):
        return

    version_pattern = r"(leaf-playground==)(\d+\.\d+\.\d+[^\s]*)"

    with open(file_path, "r") as file:
        content = file.read()

    new_content = re.sub(version_pattern, r"\g<1>" + new_version, content)

    with open(file_path, "w") as file:
        file.write(new_content)


@app.command(
    name="publish",
    help="publish project, at this time, will only copy the newest app.py and update project_config.json",
)
def publish_project(
    target: Annotated[str, typer.Argument(metavar="target_dir")],
    version_str: Annotated[str, typer.Option("--version", "-v")] = "0.1.0",
):
    dot_leaf_dir = os.path.join(target, ".leaf")

    if not os.path.exists(dot_leaf_dir):
        raise typer.BadParameter("not a leaf playground project.")

    try:
        version.parse(version_str)
    except version.InvalidVersion:
        raise typer.BadParameter(f"expect a valid version string, got {version_str}")

    with open(os.path.join(dot_leaf_dir, "project_config.json"), "r", encoding="utf-8") as f:
        project_config = json.load(f)
        old_version_str = project_config["version"]
        project_name = project_config["name"]
    if version.parse(version_str) < version.parse(old_version_str):
        raise typer.BadParameter(f"expect version string greater or equal to {old_version_str}, got {version_str}")

    # set new version
    project_config["id"] = project_config.get("id", project_name + "_" + uuid4().hex[:8])
    project_config["version"] = version_str
    project_config["leaf_version"] = leaf_version
    project_config["created_at"] = project_config.get("created_at", "unknown")
    project_config["last_update"] = datetime.utcnow().strftime("%Y-%m-%d-%H-%M-%S")

    # get and set metadata
    pkg_root = Path(os.path.join(target, project_name))
    sys.path.insert(0, pkg_root.parent.as_posix())
    scene_class: Type[Scene] = relevantly_find_subclasses(
        root_path=pkg_root.as_posix(), prefix=project_name, base_class=Scene
    )[0]
    agent_classes: List[Type[SceneAgent]] = relevantly_find_subclasses(
        root_path=os.path.join(pkg_root, "agents"), prefix=f"{project_name}.agents", base_class=SceneAgent
    )
    evaluator_classes: List[Type[MetricEvaluator]] = relevantly_find_subclasses(
        root_path=os.path.join(pkg_root, "metric_evaluators"),
        prefix=f"{project_name}.metric_evaluators",
        base_class=MetricEvaluator,
    )
    chart_classes: List[Type[Chart]] = relevantly_find_subclasses(
        root_path=os.path.join(pkg_root, "charts"), prefix=f"{project_name}.charts", base_class=Chart
    )

    agents_metadata = defaultdict(list)
    for agent_cls in agent_classes:
        if ABC in agent_cls.__bases__:
            continue
        agents_metadata[agent_cls.role_definition.name].append(
            agent_cls.get_metadata().model_dump(mode="json", by_alias=True)
        )

    project_config["metadata"] = {
        "scene_metadata": scene_class.get_metadata().model_dump(mode="json", by_alias=True),
        "agents_metadata": agents_metadata,
        "evaluators_metadata": (
            [
                evaluator_cls.get_metadata().model_dump(mode="json", by_alias=True)
                for evaluator_cls in evaluator_classes
            ]
            if evaluator_classes
            else None
        ),
        "charts_metadata": (
            [chart_cls.get_metadata().model_dump(mode="json", by_alias=True) for chart_cls in chart_classes]
            if chart_classes
            else None
        ),
    }

    with open(os.path.join(dot_leaf_dir, "project_config.json"), "w", encoding="utf-8") as f:
        json.dump(project_config, f, indent=4, ensure_ascii=False)

    app_py_file_path = os.path.join(template_dir, "{{cookiecutter.project_name}}", ".leaf", "app.py")
    shutil.copy(app_py_file_path, os.path.join(dot_leaf_dir, "app.py"))

    _replace_version_in_dockerfile(os.path.join(target, "Dockerfile"), leaf_version)

    print(f"publish new version [{version_str}]")
    raise typer.Exit()


__web_ui_version__ = "v0.4.0"
__web_ui_release_site__ = (
    "https://github.com/LLM-Evaluation-s-Always-Fatiguing/leaf-playground-webui/releases/download"
)
__web_ui_download_url__ = __web_ui_release_site__ + f"/{__web_ui_version__}/webui-{__web_ui_version__}.zip"
__web_ui_hash_url__ = __web_ui_release_site__ + f"/{__web_ui_version__}/webui-{__web_ui_version__}.zip.sha256"


def download_web_ui() -> str:
    leaf_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), ".leaf_workspace")
    tmp_dir = os.path.join(leaf_dir, "tmp")
    os.makedirs(tmp_dir, exist_ok=True)
    web_ui_dir = os.path.join(leaf_dir, "web_ui")
    os.makedirs(web_ui_dir, exist_ok=True)

    web_ui_file_name = os.path.split(__web_ui_download_url__)[-1]
    web_ui_hash_file_name = os.path.split(__web_ui_hash_url__)[-1]

    web_ui_save_dir = os.path.join(web_ui_dir, os.path.splitext(web_ui_file_name)[0])
    web_ui_file_save_path = os.path.join(tmp_dir, web_ui_file_name)

    # check hash is the same
    remote_hash = requests.get(__web_ui_hash_url__).content.decode(encoding="utf-8")
    local_hash_save_path = os.path.join(web_ui_dir, web_ui_hash_file_name)
    if os.path.exists(local_hash_save_path):
        with open(local_hash_save_path, "r", encoding="utf-8") as f:
            local_hash = f.read().strip()
        if remote_hash == local_hash and os.path.exists(web_ui_save_dir):
            return web_ui_save_dir
    with open(local_hash_save_path, "w", encoding="utf-8") as f:
        f.write(remote_hash)

    # (re-)download web ui
    if os.path.exists(web_ui_save_dir):
        shutil.rmtree(web_ui_save_dir)
    try:
        response = urlopen(__web_ui_download_url__)
        size = int(response.headers["Content-Length"])
        with open(web_ui_file_save_path, "wb") as local_file:
            with wrap_file(response, size, description="download web ui...") as remote_file:
                for chunk in remote_file:
                    local_file.write(chunk)
    except:
        # download failed, clean
        print("download failed, clean caches")
        if os.path.exists(local_hash_save_path):
            os.remove(local_hash_save_path)
        if os.path.exists(web_ui_file_save_path):
            os.remove(web_ui_file_save_path)
        raise
    with zipfile.ZipFile(web_ui_file_save_path, "r") as zip_ref:
        zip_ref.extractall(web_ui_save_dir)
    os.remove(web_ui_file_save_path)

    return web_ui_save_dir


@app.command(name="start-server", help="start a leaf-playground server, will firstly download WEB UI if necessary.")
def start_server(
    hub_dir: Annotated[str, typer.Option("--hub")] = os.getcwd(),
    port: Annotated[int, typer.Option("--port")] = 8000,
    ui_port: Annotated[int, typer.Option(default="--ui_port")] = 3000,
    dev_dir: Annotated[Optional[str], typer.Option("--dev_dir")] = None,
    web_ui_dir: Annotated[Optional[str], typer.Option("--web_ui_dir")] = None,
    no_web_ui: Annotated[Optional[bool], typer.Option("--no_web_ui")] = False,
    runtime_env: Annotated[str, typer.Option("--runtime_env", click_type=click.Choice(["local", "docker"]))] = "local",
):
    if not no_web_ui:
        if web_ui_dir and not os.path.isdir(web_ui_dir):
            raise typer.BadParameter("value of argument '--web_ui' must be a local existed directory")
        if not web_ui_dir:
            web_ui_dir = download_web_ui()
        server_url = f"http://127.0.0.1:{port}"
        subprocess.Popen(f"node {os.path.join(web_ui_dir, 'start.js')} --port={ui_port} --server={server_url}".split())

    if dev_dir:
        dev_dir = os.path.abspath(dev_dir)
        sys.path.insert(0, dev_dir)

    from .server.app import config_server, AppConfig
    from .server.utils import get_local_ip
    from .server.task.model import TaskRunTimeEnv

    if not os.path.exists(hub_dir):
        raise typer.BadParameter(f"zoo [{hub_dir}] not exist.")

    config_server(
        config=AppConfig(
            hub_dir=Path(hub_dir),
            server_port=port,
            server_host=get_local_ip(),
            runtime_env=TaskRunTimeEnv[runtime_env.upper()],
        )
    )

    import uvicorn

    uvicorn.run(
        "leaf_playground_cli.server.app:app", port=port, host="0.0.0.0", reload=bool(dev_dir), reload_dirs=dev_dir
    )


@app.command(name="version", help="currently installed leaf-playground framework version")
def get_version():
    print(f"v{leaf_version}")


@app.command(name="web-ui-version", help="bounded web ui version currently installed leaf-playground framework uses")
def get_web_ui_version():
    print(__web_ui_version__)


# TODO: add command to migrate database using Alembic


if __name__ == "__main__":
    app()



from typing import List

from .._config import _Config, _Configurable


class EvalToolConfig(_Config):
    pass


class EvalTool(_Configurable):
    config_cls = EvalToolConfig
    config: config_cls
    activated_metrics: List[str]

    def set_activated_metrics(self, metrics: List[str]):
        self.activated_metrics = metrics

    @classmethod
    def from_config(cls, config: config_cls) -> "EvalTool":
        return cls(config=config)


__all__ = ["EvalToolConfig", "EvalTool"]



import asyncio
import json
import traceback
from enum import Enum
from os import makedirs
from os.path import join
from typing import Callable, List, Literal, Optional, Type, Union
from uuid import uuid4

from pydantic import Field

from .scene import Scene
from .scene_definition import SceneConfig, SceneDefinition
from .workers import MetricEvaluator, MetricEvaluatorConfig, MetricReporter, Logger, LogHandler
from .workers.chart import Chart
from .._config import _Config
from .._type import Singleton
from ..data.log_body import SystemLogBody, SystemEvent
from ..data.message import MessagePool
from ..utils.import_util import dynamically_import_obj, DynamicObject


class SceneObjConfig(_Config):
    scene_config_data: dict = Field(default=...)
    scene_obj: DynamicObject = Field(default=...)

    def initialize_scene(self) -> Scene:
        scene_cls: Type[Scene] = dynamically_import_obj(self.scene_obj)
        scene_config_cls: Type[SceneConfig] = scene_cls.config_cls
        return scene_cls(config=scene_config_cls(**self.scene_config_data))


class ReporterObjConfig(_Config):
    charts: List[DynamicObject] = Field(default=...)

    def initialize_reporter(self, scene_definition: SceneDefinition) -> MetricReporter:
        chart_classes = []
        for chart_dynamic_obj in self.charts:
            chart_cls: Type[Chart] = dynamically_import_obj(chart_dynamic_obj)
            chart_classes.append(chart_cls)
        return MetricReporter(scene_definition=scene_definition, chart_classes=chart_classes)


class MetricEvaluatorObjConfig(_Config):
    evaluator_config_data: dict = Field(default=...)
    evaluator_obj: DynamicObject = Field(default=...)

    def initialize_evaluator(
        self, scene_config: SceneConfig, logger: Logger, reporter: MetricReporter
    ) -> MetricEvaluator:
        evaluator_cls: Type[MetricEvaluator] = dynamically_import_obj(self.evaluator_obj)
        evaluator_config_cls: Type[MetricEvaluatorConfig] = evaluator_cls.config_cls
        return evaluator_cls(
            config=evaluator_config_cls(**self.evaluator_config_data),
            scene_config=scene_config,
            logger=logger,
            reporter=reporter,
        )


class MetricEvaluatorObjsConfig(_Config):
    evaluators: List[MetricEvaluatorObjConfig] = Field(default=[])

    def initialize_evaluators(
        self, scene_config: SceneConfig, logger: Logger, reporter: MetricReporter
    ) -> List[MetricEvaluator]:
        return [
            evaluator_obj_config.initialize_evaluator(scene_config=scene_config, logger=logger, reporter=reporter)
            for evaluator_obj_config in self.evaluators
        ]


class SceneEngineState(Enum):
    PENDING = "pending"
    RUNNING = "running"
    FINISHED = "finished"
    INTERRUPTED = "interrupted"
    PAUSED = "paused"
    FAILED = "failed"


class SceneEngineStateProxy:
    def __init__(self, bound_name: str = "_state", callbacks_attr_name: str = "_state_change_callbacks"):
        self._bound_name = bound_name
        self._callbacks_attr_name = callbacks_attr_name

    def __get__(self, instance, owner):
        return getattr(instance, self._bound_name)

    def __set__(self, instance, value):
        setattr(instance, self._bound_name, value)
        for cb in getattr(instance, self._callbacks_attr_name):
            cb()


class SceneEngine(Singleton):
    state = SceneEngineStateProxy()

    def __init__(
        self,
        scene_config: SceneObjConfig,
        evaluators_config: MetricEvaluatorObjsConfig,
        reporter_config: ReporterObjConfig,
        results_dir: str,
        state_change_callbacks: List[Callable] = [],
        log_handlers: Optional[List[LogHandler]] = None,
    ):
        self._state_change_callbacks = state_change_callbacks
        self.state = SceneEngineState.PENDING
        self._id = "engine_" + uuid4().hex[:8]
        self._results_dir = results_dir

        self.message_pool = MessagePool()
        self.logger = Logger()

        if log_handlers:
            for log_handler in log_handlers:
                self.logger.registry_handler(log_handler)

        self.scene = scene_config.initialize_scene()
        self.reporter = reporter_config.initialize_reporter(scene_definition=self.scene.scene_definition)
        self.evaluators = evaluators_config.initialize_evaluators(
            scene_config=self.scene.config, logger=self.logger, reporter=self.reporter
        )

        for evaluator in self.evaluators:
            self.scene.registry_metric_evaluator(evaluator)

    @property
    def id(self) -> str:
        return self._id

    async def _wait_agents_ready(self):
        await self.scene.wait_agents_ready()

    async def run(self):
        await self._wait_agents_ready()

        self.state = SceneEngineState.RUNNING

        for evaluator in self.evaluators:
            evaluator.start()

        self.logger.add_log(SystemLogBody(system_event=SystemEvent.SIMULATION_START))
        try:
            await self.scene.run()
        except asyncio.CancelledError:
            for evaluator in self.evaluators:
                evaluator.terminate()
        except:
            traceback.print_exc()
            self.state = SceneEngineState.FAILED
            self.logger.add_log(
                SystemLogBody(system_event=SystemEvent.SIMULATION_FAILED, log_msg=traceback.format_exc())
            )
            for evaluator in self.evaluators:
                evaluator.terminate()
        else:
            self.logger.add_log(SystemLogBody(system_event=SystemEvent.SIMULATION_FINISHED))
            await asyncio.gather(*[evaluator.join() for evaluator in self.evaluators])
            self.state = SceneEngineState.FINISHED
            self.logger.add_log(SystemLogBody(system_event=SystemEvent.EVALUATION_FINISHED))
            self.logger.add_log(SystemLogBody(system_event=SystemEvent.EVERYTHING_DONE))
        finally:
            self.save(self._results_dir)

    def pause(self):
        if self.state not in [SceneEngineState.FINISHED, SceneEngineState.INTERRUPTED, SceneEngineState.FAILED]:
            self.logger.add_log(SystemLogBody(system_event=SystemEvent.SIMULATION_PAUSED))
            self.state = SceneEngineState.PAUSED
            self.scene.pause()

    def resume(self):
        if self.state not in [SceneEngineState.FINISHED, SceneEngineState.INTERRUPTED, SceneEngineState.FAILED]:
            self.logger.add_log(SystemLogBody(system_event=SystemEvent.SIMULATION_RESUME))
            self.state = SceneEngineState.RUNNING
            self.scene.resume()

    def interrupt(self):
        if self.state not in [SceneEngineState.FINISHED, SceneEngineState.INTERRUPTED, SceneEngineState.FAILED]:
            self.logger.add_log(SystemLogBody(system_event=SystemEvent.SIMULATION_INTERRUPTED))
            self.state = SceneEngineState.INTERRUPTED
            self.scene.interrupt()

    def get_scene_config(self, mode: Literal["pydantic", "dict", "json"] = "dict") -> Union[SceneConfig, dict, str]:
        if mode == "pydantic":
            return self.scene.config
        elif mode == "dict":
            return self.scene.config.model_dump(mode="json", by_alias=True)
        elif mode == "json":
            return self.scene.config.model_dump_json()
        else:
            raise ValueError(f"invalid mode {mode}")

    def get_evaluator_configs(
        self, mode: Literal["pydantic", "dict", "json"] = "dict"
    ) -> Union[List[MetricEvaluatorConfig], List[dict], str]:
        if mode == "pydantic":
            return [evaluator.config for evaluator in self.evaluators]
        elif mode == "dict":
            return [evaluator.config.model_dump(mode="json") for evaluator in self.evaluators]
        elif mode == "json":
            return json.dumps([evaluator.config.model_dump(mode="json") for evaluator in self.evaluators])
        else:
            raise ValueError(f"invalid mode {mode}")

    def save(self, save_dir: str):
        makedirs(save_dir, exist_ok=True)

        scene_config = self.get_scene_config(mode="dict")
        evaluator_configs = self.get_evaluator_configs(mode="dict")

        with open(join(save_dir, "scene_config.json"), "w", encoding="utf-8") as f:
            json.dump(scene_config, f, indent=4, ensure_ascii=False)

        with open(join(save_dir, "evaluator_configs.json"), "w", encoding="utf-8") as f:
            json.dump(evaluator_configs, f, indent=4, ensure_ascii=False)

        for exporter in self.scene.scene_definition.log_exporters:
            exporter.export(self.logger, save_dir)

        metrics_data, charts = self.reporter.generate_reports(
            scene_config=self.get_scene_config(mode="pydantic"),
            evaluator_configs=self.get_evaluator_configs(mode="pydantic"),
            logs=self.logger.logs,
        )
        metrics_data = {
            "metrics": {
                name: (
                    [each.model_dump(mode="json") for each in data]
                    if isinstance(data, list)
                    else data.model_dump(mode="json")
                )
                for name, data in metrics_data["metrics"].items()
            },
            "human_metrics": {
                name: (
                    [each.model_dump(mode="json") for each in data]
                    if isinstance(data, list)
                    else data.model_dump(mode="json")
                )
                for name, data in metrics_data["human_metrics"].items()
            },
            "merged_metrics": {
                name: (
                    [each.model_dump(mode="json") for each in data]
                    if isinstance(data, list)
                    else data.model_dump(mode="json")
                )
                for name, data in metrics_data["merged_metrics"].items()
            },
        }
        with open(join(save_dir, "metrics.json"), "w", encoding="utf-8") as f:
            json.dump(metrics_data, f, indent=4, ensure_ascii=False)

        with open(join(save_dir, "charts.json"), "w", encoding="utf-8") as f:
            json.dump(charts, f, indent=4, ensure_ascii=False)


__all__ = [
    "SceneObjConfig",
    "MetricEvaluatorObjConfig",
    "MetricEvaluatorObjsConfig",
    "ReporterObjConfig",
    "SceneEngineState",
    "SceneEngine",
]



import random
from abc import abstractmethod
from typing import Any, List, Set, Optional

from pydantic import Field

from .base import Data


class EnvironmentVariable(Data):
    """
    Environment variable of a scene that can be updated in different ways

    :param name: the environment variable's name
    :type name: str
    :param description: the environment variable's description
    :type description: str
    :param current_value: the value that the environment variable holds for current phase
    :type current_value: Any
    """

    name: str = Field(default=...)
    description: str = Field(default=...)
    current_value: Any = Field(default=...)

    @abstractmethod
    def update(self, *args, **kwargs):
        return


class RandomnessEnvironmentVariable(EnvironmentVariable):
    """
    A subtype of EnvironmentVariable whose current_value is updated by choose value from a space randomly

    :param random_space: a value space that current_value will be chosen fom
    :type random_space: Set[Any]
    :param current_value: the value that the environment variable holds for current phase
    :type current_value: Optional[Any]
    """

    random_space: Set[Any] = Field(default=...)
    current_value: Optional[Any] = Field(default=None)

    def model_post_init(self, __context: Any) -> None:
        if self.current_value is None:
            self.update()
        if self.current_value not in self.random_space:
            raise ValueError(
                f"value of 'current_value' must in 'random_space'({self.random_space}), but get {self.current_value}."
            )

    def update(self, exclude_current: bool = False) -> None:
        """
        Update current_value

        :param exclude_current: whether exclude current value from random space when choose a new value
        :return: None
        """
        choices = self.random_space
        if exclude_current:
            choices -= {self.current_value}
        self.current_value = random.choice(list(choices))


class ChainedEnvironmentVariable(EnvironmentVariable):
    """
    A subtype of EnvironmentVariable whose potential values is defined in a chain and
    will be updated in the order of candidate values in the chain.

    :param chain: a chain that contains potential values and defines their order
    :type chain: List[Any]
    :param current_value: the value that the environment variable holds for current phase
    :type current_value: Optional[Any]
    :param recurrent: whether the chain is a recurrent chain
    :type recurrent: bool
    """

    chain: List[Any] = Field(default=...)
    current_value: Optional[Any] = Field(default=None)
    recurrent: bool = Field(default=False)

    def model_post_init(self, __context: Any) -> None:
        if self.current_value is None:
            self.current_value = self.chain[0]
        if self.current_value not in self.chain:
            raise ValueError(f"value of 'current_value' must in 'chain'({self.chain}), but get {self.current_value}.")

    def update(self) -> None:
        current_position = self.chain.index(self.current_value)
        next_position = (current_position + 1) % len(self.chain)
        if current_position < next_position or self.recurrent:
            self.current_value = self.chain[next_position]


class NumericEnvironmentVariable(EnvironmentVariable):
    """
    A subtype of EnvironmentVariable whose value can be updated by using numeric operators.

    :param current_value: the value that the environment variable holds for current phase
    :type current_value: float
    :param le: the environment variable's maximum value, if is None, there is no maximum constrain
    :type le: Optional[float]
    :param ge: the environment variable's minimum value, if is None, there is no minimum constrain
    """

    current_value: float = Field(default=...)
    le: Optional[float] = Field(default=None)
    ge: Optional[float] = Field(default=None)

    def model_post_init(self, __context: Any) -> None:
        self._check_value_legality(value=self.current_value, value_name="current_value")

    def _check_value_legality(self, value: float, value_name: str = "value") -> None:
        legal = True
        if self.le is not None and value > self.le:
            legal = False
        if self.ge is not None and value < self.ge:
            legal = False
        if not legal:
            raise ValueError(f"'{value_name}'({self.current_value}) isn't legal for 'le'={self.le} and 'ge'={self.ge}")

    def update(self, new_value: float) -> None:
        self._check_value_legality(value=new_value, value_name="new_value")
        self.current_value = new_value

    def __add__(self, other: float):
        new_value = self.current_value + other
        self.update(new_value)

    def __sub__(self, other: float):
        new_value = self.current_value - other
        self.update(new_value)


class ConstantEnvironmentVariable(EnvironmentVariable):
    """
    A subtype of EnvironmentVariable whose values are unchangeable.
    """

    def update(self, *args, **kwargs):
        raise TypeError(f"{self.__class__.__name__} is unchangeable.")

    def __setattr__(self, key, value):
        if key in self.__dict__:
            raise TypeError(f"{self.__class__.__name__}.{key} is unchangeable.")


__all__ = [
    "EnvironmentVariable",
    "RandomnessEnvironmentVariable",
    "ChainedEnvironmentVariable",
    "NumericEnvironmentVariable",
    "ConstantEnvironmentVariable",
]



from typing import Literal, Optional, Dict

from jinja2 import Template
from openai.types.chat import ChatCompletionSystemMessageParam, ChatCompletionUserMessageParam
from pydantic import Field

from .base import EvalTool, EvalToolConfig
from ..ai_backend.openai import OpenAIBackend, OpenAIBackendConfig


class EvalOpenAIBackendConfig(OpenAIBackendConfig):
    model: Literal["gpt-4-1106-preview", "gpt-4"] = Field(default="gpt-4-1106-preview")


class GeneralOpenEvalToolConfig(EvalToolConfig):
    openai_backend_config: EvalOpenAIBackendConfig = Field(...)


class GeneralOpenEvalTool(EvalTool):
    config_obj = GeneralOpenEvalToolConfig
    config: config_obj

    temperature: float
    response_format: Dict[Literal["type"], Literal["text", "json"]]
    max_tokens: int

    def __init__(self, config: config_obj):
        super().__init__(config)
        self.client = OpenAIBackend.from_config(self.config.openai_backend_config).async_client
        self.temperature = 0.25
        self.response_format = {"type": "text"}
        self.max_tokens = 128

    def set_temperature(self, temperature: float):
        self.temperature = temperature

    def set_response_format(self, response_format: Literal["text", "json"]):
        self.response_format = {"type": response_format}

    def set_max_tokens(self, max_tokens: int):
        self.max_tokens = max_tokens

    async def evaluate(self, system_template: Optional[str], prompt_template: str, value_dict: dict) -> str:
        messages = []
        if system_template:
            tpl = Template(system_template)
            messages.append(ChatCompletionSystemMessageParam(role="system", content=tpl.render(value_dict)))
        tpl = Template(prompt_template)
        messages.append(ChatCompletionUserMessageParam(role="user", content=tpl.render(value_dict)))

        result = (
            (
                await self.client.chat.completions.create(
                    messages=messages,
                    model=self.config.openai_backend_config.model,
                    max_tokens=self.max_tokens,
                    response_format=self.response_format,
                    temperature=self.temperature,
                )
            )
            .choices[0]
            .message.content
        )

        return result



import asyncio
import os
import signal
import traceback
from datetime import datetime

import aiohttp
import requests
import sys
from argparse import ArgumentParser
from contextlib import asynccontextmanager

from fastapi import status, FastAPI, Depends, HTTPException, WebSocket, WebSocketDisconnect
from fastapi.responses import JSONResponse
from leaf_playground._type import Singleton
from leaf_playground.core.workers import Logger, LogHandler
from leaf_playground.core.scene_agent import HumanConnection
from leaf_playground.core.scene_engine import SceneEngine, SceneEngineState
from leaf_playground.data.log_body import LogBody, ActionLogBody
from leaf_playground.data.message import MessagePool
from leaf_playground_cli.server.task import *

sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

parser = ArgumentParser()
parser.add_argument("--id", type=str)
parser.add_argument("--port", type=int)
parser.add_argument("--host", type=str)
parser.add_argument("--secret_key", type=str)
parser.add_argument("--server_url", type=str)
parser.add_argument("--docker", action="store_true")
args = parser.parse_args()


class DBLogHandler(Singleton, LogHandler):
    def __init__(self):
        super().__init__()

        self._message_pool = MessagePool()
        self._submitted_messages = set()
        self._http_session = aiohttp.ClientSession(base_url=args.server_url)

        self._queue = asyncio.Queue()

        asyncio.ensure_future(self.db_write_loop())

    async def db_write_loop(self):
        while True:
            log_body: LogBody = await self._queue.get()
            is_update = log_body.created_at != log_body.last_update
            if not is_update:
                if isinstance(log_body, ActionLogBody):
                    message = self._message_pool.get_message_by_id(log_body.response)
                    if message.id not in self._submitted_messages:
                        self._submitted_messages.add(message.id)
                        async with self._http_session.post(
                            f"/task/{args.id}/messages/insert?secret_key={args.secret_key}",
                            headers={'Content-Type': 'application/json'},
                            json=Message.init_from_message(message, args.id).model_dump(mode="json", by_alias=True)
                        ) as resp:
                            if resp.status != 200:
                                print(f"task [{args.id}] insert message [{message.id}] to database failed.")
                                print(await resp.text())
                async with self._http_session.post(
                    f"/task/{args.id}/logs/insert?secret_key={args.secret_key}",
                    headers={'Content-Type': 'application/json'},
                    json=Log.init_from_log_body(log_body, args.id).model_dump(mode="json", by_alias=True)
                ) as resp:
                    if resp.status != 200:
                        print(f"task [{args.id}] insert log [{log_body.id}] to database failed.")
                        print(await resp.text())
            else:
                async with self._http_session.patch(
                    f"/task/{args.id}/logs/update?secret_key={args.secret_key}",
                    headers={'Content-Type': 'application/json'},
                    json=Log.init_from_log_body(log_body, args.id).model_dump(mode="json", by_alias=True)
                ) as resp:
                    if resp.status != 200:
                        print(f"task [{args.id}] update log [{log_body.id}] to database failed.")
                        print(await resp.text())

    async def notify_create(self, log_body: LogBody):
        self._queue.put_nowait(log_body)

    async def notify_update(self, log_body: LogBody):
        log_body.last_update = datetime.utcnow()
        self._queue.put_nowait(log_body)


def create_engine():
    DBLogHandler()
    try:
        resp = requests.get(f"{args.server_url}/task/{args.id}/payload")
        payload = TaskCreationPayload(**resp.json())

        if args.docker:
            results_dir = "/tmp/result"
        else:
            resp = requests.get(f"{args.server_url}/task/{args.id}/results_dir")
            results_dir = resp.json()["results_dir"]

        scene_engine = SceneEngine(
            scene_config=payload.scene_obj_config,
            evaluators_config=payload.metric_evaluator_objs_config,
            reporter_config=payload.reporter_obj_config,
            results_dir=results_dir,
            state_change_callbacks=[scene_engine_state_change_callback],
            log_handlers=[DBLogHandler.get_instance()]
        )
        asyncio.create_task(scene_engine.run())
    except:
        traceback.print_exc()
        update_task_status(SceneEngineState.FAILED.value)


def update_task_status(task_status: str):
    try:
        requests.patch(
            f"{args.server_url}/task/{args.id}/status?task_status={task_status}&secret_key={args.secret_key}",
        )
    except:
        pass


def scene_engine_state_change_callback():
    try:
        scene_engine = SceneEngine.get_instance()
    except:
        scene_engine = None
    task_status = scene_engine.state.value if scene_engine is not None else SceneEngineState.PENDING.value
    update_task_status(task_status)


class AppManager(Singleton):
    def __init__(self):
        self.shutdown_event = asyncio.Event()
        self.shutdown_task = asyncio.create_task(self.maybe_shutdown(), name="shutdown_service")

    async def maybe_shutdown(self):
        await self.shutdown_event.wait()
        await asyncio.sleep(3)
        os.kill(os.getpid(), signal.SIGTERM)


@asynccontextmanager
async def lifespan(application: FastAPI):
    create_engine()
    app_manager = AppManager()

    try:
        yield
    except:
        traceback.print_exc()

    app_manager.shutdown_task.cancel()


app = FastAPI(lifespan=lifespan)


@app.get("/hello")
async def hello() -> str:
    return "hello world!"


@app.get("/agents_connected")
async def agents_connected(scene_engine: SceneEngine = Depends(SceneEngine.get_instance)) -> JSONResponse:
    return JSONResponse(
        content={
            agent_id: agent.connected for agent_id, agent in scene_engine.scene.dynamic_agents.items()
        }
    )


@app.websocket("/ws/human/{agent_id}")
async def human_input(
    websocket: WebSocket,
    agent_id: str,
    scene_engine: SceneEngine = Depends(SceneEngine.get_instance)
):
    try:
        agent = scene_engine.scene.get_dynamic_agent(agent_id)
    except KeyError:
        await websocket.close(code=403, reason=f"agent [{agent_id}] not exists.")
        return
    if agent.connected:
        await websocket.close(code=403, reason=f"agent [{agent_id}] already connected.")
        return
    connection = HumanConnection(
        agent=agent,
        socket=websocket
    )
    await connection.connect()
    await connection.run()


@app.post("/pause")
async def pause_engine(scene_engine: SceneEngine = Depends(SceneEngine.get_instance)):
    scene_engine.pause()


@app.post("/resume")
async def resume_engine(scene_engine: SceneEngine = Depends(SceneEngine.get_instance)):
    scene_engine.resume()


@app.post("/interrupt")
async def interrupt_engine(
    scene_engine: SceneEngine = Depends(SceneEngine.get_instance),
    app_manager: AppManager = Depends(AppManager.get_instance)
):
    scene_engine.interrupt()
    app_manager.shutdown_event.set()


@app.post("/close")
async def close_engine(
    scene_engine: SceneEngine = Depends(SceneEngine.get_instance),
    app_manager: AppManager = Depends(AppManager.get_instance)
):
    if scene_engine.state not in [SceneEngineState.INTERRUPTED, SceneEngineState.FAILED, SceneEngineState.FINISHED]:
        raise HTTPException(status_code=status.HTTP_403_FORBIDDEN, detail="task not done!")
    scene_engine.save(scene_engine._results_dir)
    app_manager.shutdown_event.set()


@app.post("/save")
async def save_engine(
    scene_engine: SceneEngine = Depends(SceneEngine.get_instance)
):
    scene_engine.save(scene_engine._results_dir)


@app.post("/logs/{log_id}/record/metric/update")
async def update_metric_record(
    log_id: str,
    agent_id: str,
    record: LogEvalMetricRecord,
    logger: Logger = Depends(Logger.get_instance),
    scene_engine: SceneEngine = Depends(SceneEngine.get_instance)
):
    if not logger.is_log_exists(log_id):
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"log [{log_id}] not exist in task [{args.id}]"
        )

    data = {
        "value": record.value,
        "evaluator": "human",
        "target_agent": agent_id,
        "reason": record.reason
    }
    reporter = scene_engine.reporter
    if record.metric_name not in reporter.metric_definitions:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"expected metrics are [{list(reporter.metric_definitions.keys())}], got [{record.metric_name}]"
        )
    _, record_model = reporter.metric_definitions[record.metric_name].create_data_models()
    try:
        record_data = record_model(**data)
    except:
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail="can not parse data properly, this is mainly because the mismatch of value dtype, expected value "
                   f"dtype is {str(record_model.model_fields['value'].annotation)}"
        )
    logger.add_action_log_record(
        log_id=log_id,
        records={
            record.metric_name: data
        },
        field_name="human_eval_records"
    )
    reporter.put_human_record(record=record_data, metric_belonged_chain=record.metric_name, log_id=log_id)


@app.post("/logs/{log_id}/record/compare/update")
async def update_compare_record(
    log_id: str,
    record: LogEvalCompareRecord,
    logger: Logger = Depends(Logger.get_instance),
    scene_engine: SceneEngine = Depends(SceneEngine.get_instance)
):
    if not logger.is_log_exists(log_id):
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"log [{log_id}] not exist in task [{args.id}]"
        )

    data = {
        "value": record.value,
        "reason": record.reason,
        "evaluator": "human"
    }
    reporter = scene_engine.reporter
    if record.metric_name not in reporter.metric_definitions:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"expected metrics are [{list(reporter.metric_definitions.keys())}], got [{record.metric_name}]"
        )
    _, record_model = reporter.metric_definitions[record.metric_name].create_data_models()
    try:
        record_data = record_model(**data)
    except:
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail="can not parse data properly, this is mainly because the mismatch of value dtype, expected value "
                   f"dtype is {str(record_model.model_fields['value'].annotation)}"
        )
    logger.add_action_log_record(
        log_id=log_id,
        records={
            record.metric_name: data
        },
        field_name="human_compare_records"
    )
    reporter.put_human_record(record=record_data, metric_belonged_chain=record.metric_name, log_id=log_id)


if __name__ == "__main__":
    import uvicorn

    uvicorn.run(app, host=args.host, port=args.port)



from typing import Literal

import altair as alt
from pandas import DataFrame

from .base_vega_chart import BaseVegaChart


class SimpleBar(BaseVegaChart):
    def __init__(self, mode: Literal["percent", "value"] = "value", max_value=1.0):
        self.mode = mode
        self.max_value = max_value

    def generate(self, data: DataFrame) -> dict:
        # use the data format [{agent: gpt4, color: "#7D3C98", value: 4}, ...]

        fmt = ".2f" if self.mode == "value" else ".1%"

        chart = (
            alt.Chart(data)
            .mark_bar(cornerRadius=5, height={"band": 0.6})
            .encode(
                y=alt.X("agent:N").axis(labelAngle=0, title=None),
                x=alt.Y("value:Q").axis(title=None, format=fmt).scale(domain=(0, self.max_value)),
                color=alt.Color("color:N").scale(None).legend(None),
            )
        )

        text = chart.mark_text(align="left", dx=3).encode(text=alt.Text("value:Q", format=fmt))

        return (chart + text).to_dict()


if __name__ == "__main__":
    # This is a test
    from pandas import DataFrame
    import json

    data = DataFrame([
        {"agent": "gpt4", "color": "#7D3C98", "value": 4},
        {"agent": "gpt3", "color": "#F39C12", "value": 3},
        {"agent": "gpt2", "color": "#3498DB", "value": 2},
        {"agent": "gpt1", "color": "#E74C3C", "value": 1},
    ])

    chart = SimpleBar(mode="value", max_value=5)
    print(json.dumps(chart.generate(data)))



import asyncio
from abc import abstractmethod, ABC, ABCMeta
from inspect import signature
from sys import _getframe
from typing import Any, Dict, Optional

from pydantic import create_model, BaseModel, Field
from fastapi import WebSocket
from fastapi.websockets import WebSocketState

from .scene_definition import RoleDefinition
from .._config import _Config, _Configurable
from .._type import Immutable
from ..ai_backend.base import AIBackend, AIBackendConfig
from ..data.environment import EnvironmentVariable
from ..data.profile import Profile
from ..data.socket_data import SocketEvent
from ..utils.import_util import dynamically_import_obj, DynamicObject
from ..utils.type_util import validate_type


class _ActionHandler:
    def __init__(self, action_fn: callable, action_name: str, exec_timeout: int, executable: asyncio.Event):
        self.action_fn = action_fn
        self.action_name = action_name
        self.exec_timeout = exec_timeout

        self.executable = executable
        # this will force the same action of one instance be called once in every moment
        # thus calling the same action of one instance multiple times concurrently
        # it will still run sequentially
        self._lock = asyncio.Lock()
        self._task = None
        self._clock = None
        self._executed_seconds = 0.0

    async def _record_executed_seconds(self):
        interval = 1
        while True:
            await asyncio.sleep(interval)
            self._executed_seconds += interval

    def _reset_clock(self):
        self._clock: asyncio.Task
        if self._clock is not None:
            self._clock.cancel()
        self._clock = None

    def _reset(self):
        self._task = None
        self._reset_clock()
        self._executed_seconds = 0.0

    async def execute(self, *args, **kwargs):
        await self.executable.wait()
        try:
            async with self._lock:
                if not self._clock:
                    self._clock = asyncio.ensure_future(self._record_executed_seconds())
                self._task = asyncio.ensure_future(
                    asyncio.wait_for(
                        self.action_fn(*args, **kwargs), timeout=self.exec_timeout - self._executed_seconds
                    )
                )
                await self._task
                res = self._task.result()
                self._reset()
            return res
        except asyncio.CancelledError as e:
            if not self.executable.is_set():  # cancel will be called when agent is paused
                self._reset_clock()
                return await self.execute(*args, **kwargs)
            else:  # other case, raise cancel error
                self._reset()
                raise
        except asyncio.TimeoutError as e:
            self._reset()
            raise TimeoutError(f"action [{self.action_name}] execution exceeded the time limit: {self.exec_timeout}s.")
        except Exception as e:
            self._reset()
            raise

    @property
    def task(self):
        return self._task


class _HumanActionHandler(_ActionHandler):
    def __init__(
        self,
        human_agent: "SceneHumanAgent",
        action_fn: callable,
        action_name: str,
        exec_timeout: int,
        executable: asyncio.Event,
    ):
        super().__init__(action_fn, action_name, exec_timeout, executable)
        self.human_agent = human_agent

    async def execute(self, *args, **kwargs):
        try:
            return await super().execute(*args, **kwargs)
        except:
            self.human_agent.wait_human_input = False
            if self.human_agent.connected:
                self.human_agent.connection.notify_to_cancel()
                self.human_agent.connection.notify_human_to_not_input()
            raise


class SceneAgentMetadata(BaseModel):
    cls_name: str = Field(default=...)
    description: str = Field(default=...)
    config_schema: Optional[dict] = Field(default=...)
    obj_for_import: DynamicObject = Field(default=...)
    is_human: bool = Field(default=...)


class SceneAgentMetaClass(ABCMeta):
    def __new__(cls, name, bases, attrs, *, role_definition: RoleDefinition = None, cls_description: str = None):
        attrs["role_definition"] = Immutable(role_definition or getattr(bases[0], "role_definition", None))
        attrs["cls_description"] = Immutable(cls_description)
        attrs["obj_for_import"] = Immutable(DynamicObject(obj=name, module=_getframe(1).f_globals["__name__"]))

        new_cls = super().__new__(cls, name, bases, attrs)

        DynamicObject.bind_dynamic_obj(attrs["obj_for_import"], new_cls)

        if not validate_type(attrs["role_definition"], Immutable[Optional[RoleDefinition]]):
            raise TypeError(
                f"class [{name}]'s class attribute [role_definition] should be a [RoleDefinition] instance, "
                f"got [{type(attrs['role_definition']).__name__}] type"
            )
        if not validate_type(attrs["cls_description"], Immutable[Optional[str]]):
            raise TypeError(
                f"class [{name}]'s class attribute [cls_description] should be a [str] instance, "
                f"got [{type(attrs['cls_description']).__name__}] type"
            )

        if ABC not in bases:
            # check if those class attrs are empty when the class is not abstract
            if not new_cls.role_definition:
                raise AttributeError(
                    f"class [{name}] missing class attribute [role_definition], please specify it by "
                    f"doing like: `class {name}(role_definition=your_role_def)`, or you can also "
                    f"specify in the super class [{bases[0].__name__}]"
                )
            if not new_cls.cls_description:
                raise AttributeError(
                    f"class [{name}] missing class attribute [cls_description], please specify it by "
                    f"doing like: `class {name}(cls_description=your_cls_desc)`, where 'your_cls_desc' "
                    "is a string that introduces your agent class"
                )
            # bind the agent class to its role definition
            new_cls.role_definition._agents_cls.append(new_cls)

        if new_cls.role_definition:
            for action in new_cls.role_definition.actions:
                action_name = action.name
                action_sig = action.get_signature()

                if action_name not in attrs:
                    raise AttributeError(f"missing [{action_name}] action in class [{name}]")
                if not callable(attrs[action_name]):
                    raise TypeError(f"[{action_name}] action must be a method of class [{name}]")
                if not asyncio.iscoroutinefunction(attrs[action_name]):
                    raise TypeError(f"[{action_name}] action must be an asynchronous method of class [{name}]")
                if signature(attrs[action_name]) != action_sig:
                    raise TypeError(
                        f"expected signature of [{action_name}] action in class [{name}] is {str(action_sig)}, "
                        f"got {str(signature(attrs[action_name]))}"
                    )

        return new_cls

    def __init__(cls, name, bases, attrs, *, role_definition: RoleDefinition = None, cls_description: str = None):
        super().__init__(name, bases, attrs)

    def __setattr__(self, key, value):
        # make sure those class attributes immutable in class-wise
        if key in ["role_definition", "cls_description", "obj_for_import"] and hasattr(self, key):
            raise AttributeError(f"class attribute {key} is immutable")
        return super().__setattr__(key, value)

    def __delattr__(self, item):
        # make sure those class attributes can't be deleted
        if item in ["role_definition", "cls_description", "obj_for_import"] and hasattr(self, item):
            raise AttributeError(f"class attribute [{item}] can't be deleted")
        return super().__delattr__(item)


class SceneAgentConfig(_Config):
    profile: Profile = Field(default=...)
    chart_major_color: Optional[str] = Field(default=None, pattern=r"^#[0-9a-fA-F]{6}$")


class SceneAgent(_Configurable, ABC, metaclass=SceneAgentMetaClass):
    config_cls = SceneAgentConfig
    config: config_cls

    # class attributes initialized in metaclass
    role_definition: RoleDefinition
    cls_description: str
    obj_for_import: DynamicObject

    def __init__(self, config: config_cls):
        super().__init__(config=config)

        self._profile = self.config.profile
        self._role = self.role_definition.role_instance
        self._profile.role = self._role
        self._env_vars: Dict[str, EnvironmentVariable] = None

        self._not_paused = asyncio.Event()
        self._not_paused.set()
        self._action2handler = {}
        if ABC not in self.__class__.__bases__:
            for action in self.role_definition.actions:
                action_name = action.name
                action_handler = _ActionHandler(
                    getattr(self, action_name), action_name, action.exec_timeout, self._not_paused
                )
                setattr(self, action_name, action_handler.execute)
                self._action2handler[action_name] = action_handler

    def bind_env_vars(self, env_vars: Dict[str, EnvironmentVariable]):
        self._env_vars = env_vars

    def pause(self):
        self._not_paused.clear()
        for action_handler in self._action2handler.values():
            if action_handler.task is not None and not action_handler.task.done():
                # TODO: can't using cancel in the future if we want to support streaming the action result
                action_handler.task.cancel()

    def resume(self):
        self._not_paused.set()

    @property
    def env_vars(self) -> Dict[str, EnvironmentVariable]:
        return self._env_vars

    @property
    def profile(self):
        return self._profile

    @property
    def role(self):
        return self._role

    @property
    def id(self):
        return self.profile.id

    @property
    def name(self):
        return self.profile.name

    @property
    def role_name(self):
        return self.role.name

    @property
    def env_var(self):
        return self._env_vars

    @classmethod
    def from_config(cls, config: config_cls) -> "SceneAgent":
        return cls(config=config)

    @classmethod
    def get_metadata(cls):
        return SceneAgentMetadata(
            cls_name=cls.__name__,
            description=cls.cls_description,
            config_schema=cls.config_cls.get_json_schema(by_alias=True) if not cls.role_definition.is_static else None,
            obj_for_import=cls.obj_for_import,
            is_human=False,
        )


class SceneDynamicAgentConfig(SceneAgentConfig):
    def model_post_init(self, __context) -> None:
        pass


class SceneDynamicAgent(SceneAgent, ABC):
    config_cls = SceneDynamicAgentConfig
    config: config_cls

    def __init__(self, config: config_cls):
        super().__init__(config=config)

        self.connected = False

    @abstractmethod
    def connect(self, *args, **kwargs):
        pass

    @abstractmethod
    def disconnect(self, *args, **kwargs):
        pass


class SceneAIAgentConfig(SceneDynamicAgentConfig):
    ai_backend_config: AIBackendConfig = Field(default=...)
    ai_backend_obj: DynamicObject = Field(default=..., exclude=True)

    def model_post_init(self, __context) -> None:
        self.valid(self.ai_backend_config, self.ai_backend_obj)

    def create_backend_instance(self) -> AIBackend:
        obj = dynamically_import_obj(self.ai_backend_obj)
        return obj.from_config(config=self.ai_backend_config)

    @staticmethod
    def valid(ai_backend_config: AIBackendConfig, ai_backend_obj: DynamicObject):
        ai_backend_cls = dynamically_import_obj(ai_backend_obj)
        if not issubclass(ai_backend_cls, AIBackend):
            raise TypeError(f"ai_backend_obj {ai_backend_obj.obj} should be a subclass of AIBackend")
        if not isinstance(ai_backend_config, ai_backend_cls.config_cls):
            raise TypeError(f"ai_backend_config should be an instance of {ai_backend_cls.config_cls.__name__}")


class SceneAIAgent(SceneDynamicAgent, ABC):
    config_cls = SceneAIAgentConfig
    config: config_cls

    def __init__(self, config: config_cls):
        super().__init__(config=config)

        self.backend = self.config.create_backend_instance()
        self.connected = True

    def connect(self):
        pass

    def disconnect(self):
        pass


class HumanConnection:
    def __init__(self, agent: "SceneHumanAgent", socket: WebSocket):
        self.agent = agent
        self.socket = socket

        self.events = asyncio.Queue()
        self.state = WebSocketState.CONNECTING

        self.cancel_event = asyncio.Event()

    async def connect(self):
        await self.socket.accept()
        self.agent.connect(self)
        self.state = WebSocketState.CONNECTED
        if self.agent.wait_human_input:
            self.notify_human_to_input()
        else:
            self.notify_human_to_not_input()

    def disconnect(self):
        self.agent.disconnect()
        self.state = WebSocketState.DISCONNECTED

    def notify_human_to_input(self):
        self.events.put_nowait(SocketEvent(event="wait_human_input"))

    def notify_human_to_not_input(self):
        self.events.put_nowait(SocketEvent(event="disable_human_input"))

    def notify_to_cancel(self):
        self.cancel_event.set()

    async def _keep_alive(self):
        try:
            while True:
                await self.socket.send_json(SocketEvent(event="heart_beat").model_dump(mode="json"))
                await asyncio.sleep(0.1)
        except:
            self.disconnect()
            return

    async def _send_socket_event(self):
        while True:
            if self.state == WebSocketState.DISCONNECTED:
                break
            try:
                event: SocketEvent = await self.events.get()
                await self.socket.send_json(event.model_dump(mode="json"))
            except:
                continue

    async def _receive_human_input(self):
        while True:
            if self.state == WebSocketState.DISCONNECTED:
                break
            try:
                fut = asyncio.ensure_future(self.socket.receive_text(), loop=asyncio.get_running_loop())
                while not fut.done():
                    if self.cancel_event.is_set():
                        fut.cancel()
                        self.cancel_event.clear()
                        break
                    await asyncio.sleep(0.01)
                try:
                    await fut
                except asyncio.CancelledError:
                    continue
                if fut.done():
                    self.agent.human_input = fut.result()
            except:
                continue

    async def run(self):
        await asyncio.gather(*[self._send_socket_event(), self._receive_human_input(), self._keep_alive()])


class SceneHumanAgentConfig(SceneDynamicAgentConfig):
    pass


class SceneHumanAgent(SceneDynamicAgent, ABC):
    config_cls = SceneHumanAgentConfig
    config: config_cls

    def __init__(self, config: config_cls):
        super().__init__(config=config)

        self.connection: HumanConnection = None
        self.human_input = None
        self.wait_human_input = False

        if ABC not in self.__class__.__bases__:
            for action in self.role_definition.actions:
                action_name = action.name
                action_handler = _HumanActionHandler(
                    self,
                    self._action2handler[action_name].action_fn,
                    action_name,
                    action.exec_timeout,
                    self._not_paused,
                )
                setattr(self, action_name, action_handler.execute)
                self._action2handler[action_name] = action_handler

    def connect(self, connection: HumanConnection):
        self.connection = connection
        self.connected = True

    def disconnect(self):
        self.connection = None
        self.connected = False

    async def wait_human_text_input(self) -> Optional[str]:
        if not self.connected:
            return None
        self.wait_human_input = True
        self.connection.notify_human_to_input()
        while not self.human_input:
            await asyncio.sleep(0.1)
        self.wait_human_input = False
        if self.connected:
            self.connection.notify_human_to_not_input()
        human_input = self.human_input
        self.human_input = None
        return human_input

    async def wait_human_image_input(self, *args, **kwargs):
        raise NotImplementedError()  # TODO: impl

    @classmethod
    def get_metadata(cls):
        metadata = super().get_metadata()
        metadata.is_human = True
        return metadata


class SceneStaticAgentConfig(SceneAgentConfig):
    def model_post_init(self, __context: Any) -> None:
        fields = self.model_fields_set
        if fields - {"profile", "chart_major_color"}:
            raise ValueError(f"{self.__class__.__name__} requires profile and chart_major_color only, got {fields}")

    @classmethod
    def create_config_model(cls, role_definition: RoleDefinition) -> "SceneStaticAgentConfig":
        model_name = "".join([each.capitalize() for each in role_definition.name.split("_")]) + "Config"
        module = _getframe(1).f_code.co_filename
        fields = {
            "profile": (Profile, Field(default=Profile(name=role_definition.name), frozen=True, exclude=True)),
            "chart_major_color": (Optional[str], Field(default=None, pattern=r"^#[0-9a-fA-F]{6}$", exclude=True)),
        }
        return create_model(__model_name=model_name, __module__=module, __base__=cls, **fields)


class SceneStaticAgent(SceneAgent, ABC):
    config_cls = SceneStaticAgentConfig
    config: config_cls

    def __init__(self, config: config_cls):
        super().__init__(config=config)


__all__ = [
    "SceneAgentMetadata",
    "SceneAgentConfig",
    "SceneAgent",
    "SceneStaticAgentConfig",
    "SceneStaticAgent",
    "SceneDynamicAgentConfig",
    "SceneDynamicAgent",
    "SceneAIAgentConfig",
    "SceneAIAgent",
    "SceneHumanAgentConfig",
    "SceneHumanAgent",
]



import asyncio
from datetime import datetime
from enum import Enum
from typing import List

from sqlmodel import SQLModel, Field, Column, Relationship, DateTime, JSON

from leaf_playground.core.scene_engine import SceneEngineState
from leaf_playground.data.log_body import LogBody, LogType, ActionLogBody, SystemLogBody
from leaf_playground.data.message import Message as LEAFMessage
from leaf_playground.utils.import_util import dynamically_import_obj, DynamicObject


class TaskRunTimeEnv(Enum):
    LOCAL = "local"
    DOCKER = "docker"


class TaskDBLifeCycle(Enum):
    LIVING = "living"
    DELETING = "deleting"


class Task(SQLModel):
    id: str = Field(default=...)
    project_id: str = Field(default=...)
    port: int = Field(default=...)
    host: str = Field(default="http://127.0.0.1")
    payload: str = Field(default=...)
    status: str = Field(default=SceneEngineState.PENDING.value)
    runtime_env: TaskRunTimeEnv = Field(default=TaskRunTimeEnv.LOCAL)
    created_at: datetime = Field(default_factory=datetime.utcnow)
    results_dir: str = Field(default=...)

    @property
    def base_http_url(self) -> str:
        return f"http://{self.host}:{self.port}"

    @property
    def base_ws_url(self) -> str:
        return f"ws://{self.host}:{self.port}"


class TaskTable(Task, table=True):
    __tablename__ = "task"

    id: str = Field(default=..., primary_key=True, index=True)
    project_id: str = Field(default=..., index=True)
    status: str = Field(default=SceneEngineState.PENDING.value, index=True)
    runtime_env: TaskRunTimeEnv = Field(default=TaskRunTimeEnv.LOCAL, index=True)
    created_at: datetime = Field(default_factory=datetime.utcnow, sa_column=Column(DateTime, index=True))

    logs: List["LogTable"] = Relationship(back_populates="task")
    messages: List["MessageTable"] = Relationship(back_populates="task")

    live_cycle: TaskDBLifeCycle = Field(default=TaskDBLifeCycle.LIVING)

    def to_task(self) -> Task:
        return Task(**self.model_dump())

    @classmethod
    def from_task(cls, task: Task) -> "TaskTable":
        return cls(**task.model_dump())


class Log(SQLModel):
    __tablename__ = "log"

    id: str = Field(default=...)
    log_type: LogType = Field(default=...)
    log_msg: str = Field(default=...)
    created_at: datetime = Field(default=...)
    last_update: datetime = Field(default=...)
    db_last_update: datetime = Field(default_factory=datetime.utcnow)
    tid: str = Field(default=...)
    data: dict = Field(default=...)

    @classmethod
    def init_from_log_body(cls, log: LogBody, tid: str) -> "Log":
        log_dict = {"tid": tid, "data": {}}
        for k, v in log.model_dump(mode="json", by_alias=True).items():
            if k in ["id", "log_type", "log_msg", "created_at", "last_update"]:
                log_dict[k] = getattr(log, k)
            else:
                log_dict["data"][k] = v
        return cls(**log_dict)


class LogTable(Log, table=True):
    id: str = Field(default=..., primary_key=True, index=True)
    created_at: datetime = Field(default=..., sa_column=Column(DateTime, index=True))
    last_update: datetime = Field(default=..., sa_column=Column(DateTime, index=True))
    db_last_update: datetime = Field(default_factory=datetime.utcnow, sa_column=Column(DateTime, index=True))
    tid: str = Field(default=..., foreign_key="task.id", index=True)
    data: dict = Field(default=..., sa_column=Column(JSON))

    task: TaskTable = Relationship(back_populates="logs")

    def to_log(self) -> Log:
        return Log(**self.model_dump())

    @classmethod
    def from_log(cls, log: Log) -> "LogTable":
        return cls(**log.model_dump())


class Message(SQLModel):
    __tablename__ = "message"

    id: str = Field(default=...)
    tid: str = Field(default=...)
    data: dict = Field(default=...)
    obj: dict = Field(default=...)

    @classmethod
    def init_from_message(cls, message: LEAFMessage, tid: str) -> "Message":
        msg_dict = {
            "id": message.id,
            "tid": tid,
            "data": message.model_dump(mode="json", by_alias=True),
            "obj": DynamicObject.create_dynamic_obj(message.__class__).model_dump(mode="json", by_alias=True),
        }

        return cls(**msg_dict)


class MessageTable(Message, table=True):
    id: str = Field(default=..., primary_key=True)
    tid: str = Field(default=..., foreign_key="task.id", index=True)
    data: dict = Field(default=..., sa_column=Column(JSON))
    obj: dict = Field(default=..., sa_column=Column(JSON))

    task: TaskTable = Relationship(back_populates="messages")

    def to_message(self) -> Message:
        return Message(**self.model_dump())

    @classmethod
    def from_message(cls, message: Message) -> "MessageTable":
        return cls(**message.model_dump())


__all__ = ["TaskRunTimeEnv", "TaskDBLifeCycle", "Task", "TaskTable", "Log", "LogTable", "Message", "MessageTable"]



from .._config import _Config, _Configurable


class AIBackendConfig(_Config):
    pass


class AIBackend(_Configurable):
    config_cls = AIBackendConfig
    config: config_cls

    @classmethod
    def from_config(cls, config: config_cls) -> "AIBackend":
        return cls(config=config)


__all__ = ["AIBackendConfig", "AIBackend"]



import asyncio
from functools import wraps, partial
from typing import Any, Callable, Type

from pydantic import BaseModel, ConfigDict, ValidationError


def validate_type(var: Any, expect_type: Type) -> bool:
    class Model(BaseModel):
        model_config = ConfigDict(strict=True, arbitrary_types_allowed=True)

        value: expect_type

    try:
        Model(value=var)
        return True
    except ValidationError:
        return False


class MethodOutputsTypeChecker:
    def _type_check_wrap(self, func: Callable = None, *, outputs_type: Type = None):
        if func is None:
            return partial(self._type_check_wrap, outputs_type=outputs_type)

        def validate(outputs: Any):
            if outputs_type is None and outputs is not None:
                raise TypeError(f"expected outputs of function {func.__name__} is None, got {outputs}")
            elif not validate_type(outputs, outputs_type):
                raise TypeError(
                    f"expected outputs type/annotation of function {func.__name__} is {outputs_type}, "
                    f"got an invalid outputs {outputs} with type {type(outputs)}"
                )
            return outputs

        @wraps(func)
        def wrap(*args, **kwargs):
            result = func(*args, **kwargs)
            return validate(result)

        @wraps(func)
        async def awrap(*args, **kwargs):
            result = await func(*args, **kwargs)
            return validate(result)

        return awrap if asyncio.iscoroutinefunction(func) else wrap

    def __init__(self, func: Callable = None, *, outputs_type: Type = None):
        self._wrap = self._type_check_wrap(func=func, outputs_type=outputs_type)
        self._func = func

    def __call__(self, *args, **kwargs):
        if not self._func:
            self._func = args[0]
            self._wrap = self._wrap(func=self._func)
            return self
        return self._wrap(*args, **kwargs)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        pass


__all__ = ["validate_type", "MethodOutputsTypeChecker"]



import asyncio
import importlib
import multiprocessing
import os
import pickle
import time
import traceback
from abc import abstractmethod, ABC, ABCMeta
from collections import defaultdict
from enum import Enum
from multiprocessing import Manager, Process
from queue import Empty
from sys import _getframe
from types import new_class
from typing import Any, Dict, List, Optional, Type, Union
from uuid import uuid4

from pydantic import BaseModel, Field

from .logger import Logger
from ..scene_definition import CompareMetricDefinition, MetricDefinition, VALUE_DETYPE_2_DEFAULT_VALUE, SceneConfig
from ..._config import _Config, _Configurable
from ..._type import Immutable
from ...data.log_body import ActionLogBody
from ...data.media import Media
from ...data.message import Message
from ...utils.import_util import DynamicObject
from ...utils.type_util import validate_type


class RecordOutput(BaseModel):
    record_value: Any = Field(default=...)
    reason: Optional[str] = Field(default=None)
    misc: Optional[dict] = Field(default=None)


class CompareOutput(BaseModel):
    compare_result: Any = Field(default=...)
    reason: Optional[str] = Field(default=None)
    misc: Optional[dict] = Field(default=None)


_MetricName = str
_MetricRecordValue = Any


class MetricEvaluatorProxy(Process):
    def __init__(
        self,
        config_cls: Type["MetricEvaluatorConfig"],
        config_data: dict,
        record_metrics: List[_MetricName],
        compare_metrics: List[_MetricName],
        manager: Manager,
        queue: multiprocessing.Queue,
        result_cache: dict,
    ):
        super().__init__(daemon=True)

        self._queue = queue
        self._result_cache = result_cache
        self._can_stop = manager.Value("b", False)
        self._state = manager.Value("c", MetricEvaluatorState.PENDING.value.encode("utf-8"))

        self._config_cls = config_cls
        self._config_data = config_data
        self._record_metrics = record_metrics
        self._compare_metrics = compare_metrics

    def __init_subclass__(cls, _init_evaluator, _record, _compare, **kwargs):
        super().__init_subclass__(**kwargs)
        cls._init_evaluator = _init_evaluator
        cls._record = _record
        cls._compare = _compare

    @property
    def can_stop(self):
        return self._can_stop.value

    @can_stop.setter
    def can_stop(self, can_stop: bool):
        self._can_stop.value = can_stop

    @property
    def result_cache(self):
        return self._result_cache

    @property
    def queue(self):
        return self._queue

    @property
    def state(self) -> "MetricEvaluatorState":
        return MetricEvaluatorState(self._state.value.decode("utf-8"))

    def run(self):
        self._state.value = MetricEvaluatorState.INITIALIZING.value.encode("utf-8")
        try:
            evaluator = self._init_evaluator(
                self._config_cls(**self._config_data), self._record_metrics, self._compare_metrics
            )
        except:
            traceback.print_exc()
            self._state.value = MetricEvaluatorState.INIT_FAILED.value.encode("utf-8")
            return
        loop = asyncio.new_event_loop()
        self._state.value = MetricEvaluatorState.RUNNING.value.encode("utf-8")

        while self._state.value == MetricEvaluatorState.RUNNING.value.encode("utf-8"):
            if self._queue.empty() and self._can_stop.value:
                self._state.value = MetricEvaluatorState.FINISHED.value.encode("utf-8")
                break

            try:
                response, references, ground_truth, kwargs, is_compare, id_ = self._queue.get_nowait()
            except Empty:
                time.sleep(0.001)
                continue
            except:
                traceback.print_exc()
                self._state.value = MetricEvaluatorState.RUN_FAILED.value.encode("utf-8")
                break
            try:
                response = pickle.loads(response)
                references = pickle.loads(references)
                ground_truth = pickle.loads(ground_truth)
                kwargs = pickle.loads(kwargs)
                if not is_compare:
                    output = loop.run_until_complete(
                        self._record(response, references, ground_truth, evaluator, **kwargs)
                    )
                else:
                    output = loop.run_until_complete(
                        self._compare(response, references, ground_truth, evaluator, **kwargs)
                    )
            except:
                traceback.print_exc()
                output = {}
            self._result_cache[id_] = {k: v.model_dump(mode="json", by_alias=True) for k, v in output.items()}

    def terminate(self):
        super().terminate()
        self._state.value = MetricEvaluatorState.TERMINATED.value.encode("utf-8")


# TODO: We hope that the MetricEvaluator class can be inherited through composition
#  (or at least subcls can extend defs), however, using meta class makes this tough


class MetricEvaluatorMetaClass(ABCMeta):
    def __new__(
        cls,
        name,
        bases,
        attrs,
        *,
        metric_definitions: List[Union[CompareMetricDefinition, MetricDefinition]] = None,
        cls_description: str = None,
    ):
        # create proxy class on the fly
        evaluator_proxy_class = new_class(
            name=f"{name}Proxy",
            bases=(MetricEvaluatorProxy,),
            kwds={
                "_init_evaluator": attrs["_init_evaluator"],
                "_record": attrs["_record"],
                "_compare": attrs["_compare"],
            },
        )
        evaluator_proxy_class.__module__ = _getframe(1).f_globals["__name__"]
        setattr(
            importlib.import_module(evaluator_proxy_class.__module__),
            evaluator_proxy_class.__name__,
            evaluator_proxy_class,
        )

        attrs["metric_definitions"] = Immutable(metric_definitions or getattr(bases[0], "metric_definitions", None))
        attrs["cls_description"] = Immutable(cls_description or "")
        attrs["evaluator_proxy_class"] = Immutable(evaluator_proxy_class)
        attrs["obj_for_import"] = Immutable(DynamicObject(obj=name, module=_getframe(1).f_globals["__name__"]))

        new_cls = super().__new__(cls, name, bases, attrs)

        DynamicObject.bind_dynamic_obj(attrs["obj_for_import"], new_cls)

        if not validate_type(attrs["metric_definitions"], Immutable[Optional[List[MetricDefinition]]]):
            raise TypeError(
                f"class [{name}]'s class attribute [metric_definitions] should be a List[MetricDefinition]"
            )
        if not validate_type(attrs["cls_description"], Immutable[Optional[str]]):
            raise TypeError(
                f"class [{name}]'s class attribute [cls_description] should be a [str] instance, "
                f"got [{type(attrs['cls_description']).__name__}] type"
            )
        if not validate_type(attrs["evaluator_proxy_class"], Immutable[Optional[MetricEvaluatorProxy]]):
            raise TypeError(
                f"class [{name}]'s class attribute [evaluator_proxy_class] should be a subclass of "
                "[MetricEvaluatorProxy]"
            )

        if ABC not in bases:
            # check if those class attrs are empty when the class is not abstract
            if not new_cls.metric_definitions:
                raise AttributeError(
                    f"class [{name}] missing class attribute [metric_definitions], please specify it by "
                    f"doing like: `Class {name}(metric_definitions=your_metric_defs)`, or you can also "
                    f"specify in the super class [{bases[0].__name__}]"
                )
            if not new_cls.cls_description:
                raise AttributeError(
                    f"class [{name}] missing class attribute [cls_description], please specify it by "
                    f"doing like: `class {name}(cls_description=your_cls_desc)`, where 'your_cls_desc' "
                    "is a string that introduces your evaluator class"
                )
            if not new_cls.evaluator_proxy_class:
                raise AttributeError(
                    f"class [{name}] missing class attribute [cls_description], please specify it by "
                    f"doing like: `class {name}(evaluator_proxy_class=your_evaluator_proxy_class)`, where "
                    "'your_evaluator_proxy_class' is the proxy class(a Process) of your evaluator class"
                )

        return new_cls

    def __init__(
        cls,
        name,
        bases,
        attrs,
        *,
        metric_definitions: List[Union[CompareMetricDefinition, MetricDefinition]] = None,
        cls_description: str = None,
    ):
        super().__init__(name, bases, attrs)

    def __setattr__(self, key, value):
        # make sure those class attributes immutable in class-wise
        if key in ["metric_definitions", "cls_description", "obj_for_import"] and hasattr(self, key):
            raise AttributeError(f"class attribute {key} is immutable")
        return super().__setattr__(key, value)

    def __delattr__(self, item):
        # make sure those class attributes can't be deleted
        if item in ["metric_definitions", "cls_description", "obj_for_import"] and hasattr(self, item):
            raise AttributeError(f"class attribute [{item}] can't be deleted")
        return super().__delattr__(item)


class MetricEvaluatorMetadata(BaseModel):
    cls_name: str = Field(default=...)
    description: str = Field(default=...)
    config_schema: dict = Field(default=...)
    obj_for_import: DynamicObject = Field(default=...)
    metrics: List[str] = Field(default=...)


class MetricEvaluatorState(Enum):
    PENDING = "pending"
    INITIALIZING = "initializing"
    INIT_FAILED = "init failed"
    RUNNING = "running"
    RUN_FAILED = "run failed"
    FINISHED = "finished"
    TERMINATED = "terminated"


class MetricEvaluatorConfig(_Config):
    max_concurrency: int = Field(default=1, gt=0, lt=2 * os.cpu_count())
    non_ignored_message_type: Optional[List[Type[Message]]] = Field(default=None, exclude=True)


class MetricEvaluator(_Configurable, ABC, metaclass=MetricEvaluatorMetaClass):
    config_cls: Type[MetricEvaluatorConfig] = MetricEvaluatorConfig
    config: config_cls

    # class attrs that are immutable
    metric_definitions: List[Union[CompareMetricDefinition, MetricDefinition]]
    cls_description: str
    evaluator_proxy_class: Type[MetricEvaluatorProxy]
    obj_for_import: DynamicObject

    def __init__(
        self,
        config: config_cls,
        scene_config: SceneConfig,
        logger: Logger,
        reporter: "leaf_playground.core.workers.MetricReporter",
    ):
        super().__init__(config=config)

        self.config_data = self.config.model_dump(mode="json", by_alias=True)

        metric_name2def = {metric_def.belonged_chain: metric_def for metric_def in self.metric_definitions}
        metric_name2conf = {
            metric_def.belonged_chain: scene_config.get_metric_config(metric_def.belonged_chain)
            for metric_def in self.metric_definitions
        }
        self.metric_name2metric_defs = {
            metric_name: metric_def
            for metric_name, metric_def in metric_name2def.items()
            if metric_name2conf[metric_name].enable
        }
        self.resp_msg_type2metric_defs = defaultdict(list)
        for metric in self.metric_name2metric_defs.values():
            self.resp_msg_type2metric_defs[metric.expect_resp_msg_type].append(metric)

        self.metrics_for_record = [
            metric_def.belonged_chain
            for metric_def in self.metric_name2metric_defs.values()
            if not metric_def.is_comparison
        ]
        self.metrics_for_compare = [
            metric_def.belonged_chain
            for metric_def in self.metric_name2metric_defs.values()
            if metric_def.is_comparison
        ]

        self.logger = logger
        self.reporter = reporter

        manager = Manager()
        self.queue = manager.Queue()
        self.result_cache = manager.dict()

        self.proxys = [
            self.evaluator_proxy_class(
                config_cls=self.config_cls,
                config_data=self.config_data,
                record_metrics=self.metrics_for_record,
                compare_metrics=self.metrics_for_compare,
                manager=manager,
                queue=self.queue,
                result_cache=self.result_cache,
            )
            for _ in range(self.config.max_concurrency)
        ]

    @staticmethod
    @abstractmethod
    def _init_evaluator(
        config: MetricEvaluatorConfig, record_metrics: List[_MetricName], compare_metrics: List[_MetricName]
    ) -> Any:
        pass

    @staticmethod
    @abstractmethod
    async def _record(
        response: Message, references: Optional[List[Message]], ground_truth: Optional[Media], evaluator: Any, **kwargs
    ) -> Dict[_MetricName, RecordOutput]:
        pass

    @staticmethod
    @abstractmethod
    async def _compare(
        response: Message, references: Optional[List[Message]], ground_truth: Optional[Media], evaluator: Any, **kwargs
    ) -> Dict[_MetricName, CompareOutput]:
        pass

    def notify_to_record(self, log: ActionLogBody):
        asyncio.ensure_future(self.record(log))

    def notify_to_compare(self, log: ActionLogBody):
        asyncio.ensure_future(self.compare(log))

    def notify_can_stop(self):
        for proxy in self.proxys:
            proxy.can_stop = True

    def start(self):
        for proxy in self.proxys:
            proxy.start()

    def terminate(self):
        for proxy in self.proxys:
            proxy.terminate()

    async def join(self):
        for proxy in self.proxys:
            while proxy.state not in [
                MetricEvaluatorState.FINISHED,
                MetricEvaluatorState.TERMINATED,
                MetricEvaluatorState.INIT_FAILED,
                MetricEvaluatorState.RUN_FAILED,
            ]:
                await asyncio.sleep(0.1)
            else:
                print(f"evaluator {proxy.name} finished")

    async def _wait_result(self, log: ActionLogBody, is_compare: bool = False):
        id_ = uuid4()
        response = self.logger.message_pool.get_message_by_id(log.response)
        references = [self.logger.message_pool.get_message_by_id(ref) for ref in (log.references or [])] or None
        kwargs = log.model_dump(
            mode="json",
            exclude={
                "log_type",
                "response",
                "references",
                "ground_truth",
                "eval_records",
                "compare_records",
                "human_eval_records",
                "human_compare_records",
            },
        )
        self.queue.put_nowait((
            pickle.dumps(response),
            pickle.dumps(references),
            pickle.dumps(log.ground_truth),
            pickle.dumps(kwargs),
            is_compare,
            id_,
        ))
        while id_ not in self.result_cache:
            await asyncio.sleep(0.1)  # sleep longer to let scene's main process have more CPU time slice
        return {k: (CompareOutput if is_compare else RecordOutput)(**v) for k, v in self.result_cache.pop(id_).items()}

    async def record(self, log: ActionLogBody) -> None:
        response = self.logger.message_pool.get_message_by_id(log.response)
        resp_type = type(response)
        if resp_type not in list(self.resp_msg_type2metric_defs.keys()) + (self.config.non_ignored_message_type or []):
            return
        target_agent = response.sender_id
        # this may very slow
        record_results = await self._wait_result(log)
        records = {}
        for metric_name, record_output in record_results.items():
            if metric_name not in self.metrics_for_record:
                continue

            record_value = record_output.record_value
            reason = record_output.reason
            misc = record_output.misc

            metric_def = self.metric_name2metric_defs[metric_name]
            expect_dtype = metric_def.record_value_dtype

            # validate value dtype
            if not validate_type(record_value, VALUE_DETYPE_2_DEFAULT_VALUE[expect_dtype]):
                continue
            # save record data
            _, record_data_model = metric_def.create_data_models()
            record_data = record_data_model(
                value=record_value,
                reason=reason,
                misc=misc,
                target_agent=target_agent,
                evaluator=self.__class__.__name__,
            )
            self.reporter.put_record(record_data, metric_def.belonged_chain, log.id)
            records[metric_name] = record_data.model_dump(mode="json")
        self.logger.add_action_log_record(log_id=log.id, records=records, field_name="eval_records")

    async def compare(self, log: ActionLogBody) -> None:
        resp_type = type(self.logger.message_pool.get_message_by_id(log.response))
        if resp_type not in self.resp_msg_type2metric_defs:
            return

        # this may very, very slow
        compare_results = await self._wait_result(log, is_compare=True)
        records = {}
        for metric_name, compare_output in compare_results.items():
            if metric_name not in self.metrics_for_compare:
                continue

            compare_result = compare_output.compare_result
            reason = compare_output.reason
            misc = compare_output.misc

            metric_def = self.metric_name2metric_defs[metric_name]

            # validate value dtype
            if not validate_type(compare_result, List[List[str]]):
                continue
            # save record data
            _, record_data_model = metric_def.create_data_models()
            record_data = record_data_model(
                value=compare_result, reason=reason, misc=misc, evaluator=self.__class__.__name__
            )
            self.reporter.put_record(record_data, metric_def.belonged_chain, log.id)
            records[metric_name] = record_data.model_dump(mode="json")
        self.logger.add_action_log_record(log_id=log.id, records=records, field_name="compare_records")

    @classmethod
    def get_metadata(cls) -> MetricEvaluatorMetadata:
        return MetricEvaluatorMetadata(
            cls_name=cls.__name__,
            description=cls.cls_description,
            config_schema=cls.config_cls.get_json_schema(by_alias=True),
            obj_for_import=cls.obj_for_import,
            metrics=[metric_def.belonged_chain for metric_def in cls.metric_definitions],
        )

    @classmethod
    def from_config(cls, config: config_cls) -> "MetricEvaluator":
        raise NotImplementedError()

    @classmethod
    def from_config_file(cls, file_path: str) -> "MetricEvaluator":
        raise NotImplementedError()

    def save_config(self, file_path: str):
        raise NotImplementedError()


__all__ = [
    "RecordOutput",
    "CompareOutput",
    "MetricEvaluatorConfig",
    "MetricEvaluatorProxy",
    "MetricEvaluator",
    "MetricEvaluatorMetadata",
]
